# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
A class hierarchy for pixel masks.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import numpy
import astropy.constants

from .bitmask import BitMask
from ..par.artifactdb import ArtifactDB
from ..par.emissionlinedb import EmissionLineDB

from matplotlib import pyplot


class PixelMask:
    """
    Base class for a general 1D or 2D pixel mask.

    Attributes:
        shape (tuple): Shape of the array for the mask.
    """
    def __init__(self):
        self.shape = None


    def _set_shape(self, x, ny=None):
        """
        Set :attr:`shape` of the object

        Args:
            x (numpy.ndarray): Vector with the x-coordinates
            ny (int): (**Optional**) Size of the second dimension.
                Default is that there is no second dimension.
        """
        self.shape = (len(x),) if ny is None else (ny,len(x))


    def _empty_mask(self, x, ny=None):
        """
        Return an empty mask with the correct shape.

        Args:
            x (numpy.ndarray): Coordinate vector
            ny (int): (**Optional**) Size of the second dimension.
                Default is that there is only one dimension.

        Returns:
            numpy.ndarray : Boolean mask of the correct shape.
        """
        self._set_shape(x, ny=ny)
        return numpy.full(self.shape, False, dtype=bool)


    def _mask_coordinate_ranges(self, x, rng, ny=None):
        """
        Flag any x coordinates between a set of range limits as True.
        The mask is repeated in the second dimension, if requested.

        Args:
            x (numpy.ndarray): Coordinate vector
            rng (list, numpy.ndarray): (List of) Coordinate ranges that
                should be masked.
            ny (int): (**Optional**) Size of the second dimension.
                Default is that there is only one dimension.

        Returns:
            numpy.ndarray : Boolean mask of the correct shape.
        """
        self._set_shape(x, ny=ny)
        _rng = numpy.atleast_2d(rng)

        # Account for any undefined limits
        _rng[numpy.equal(_rng[:,0], None),0] = x[0]-1
        _rng[numpy.equal(_rng[:,1], None),1] = x[-1]+1

        # Generate the mask
        mask = numpy.any( numpy.array([ numpy.logical_and(x>l, x<u) \
                                        for l,u in zip(_rng[:,0], _rng[:,1])]), axis=0)
        return mask if len(self.shape) == 1 else numpy.array([mask]*self.shape[0])
        

class SpectralPixelMask(PixelMask):
    """
    Container that produces a mask for the stellar continuum based on a
    set of emission lines and artifacts.

    Args:
        artdb (:class:`mangadap.proc.artifactdb.ArtifactDB`):
            (**Optional**) Database with the list of artifacts to mask.
        emldb (:class:`mangadap.proc.emissionlinedb.EmissionLineDB`):
            (**Optional**) Database with the list of emission lines to
            mask.
        waverange (numpy.ndarray): (**Optional**) Any pixels **outside**
            this wavelength range are masked.

    Attributes:
        artdb (:class:`mangadap.proc.artifactdb.ArtifactDB`):
            Database with the list of artifacts to mask.
        emldb (:class:`mangadap.proc.emissionlinedb.EmissionLineDB`):
            Database with the list of emission lines to
            mask.
        waverange (numpy.ndarray): Any pixels **outside**
            this wavelength range are masked.
    """
    def __init__(self, artdb=None, emldb=None, waverange=None, nsig=None):
        if artdb is not None and not isinstance(artdb, ArtifactDB):
            raise TypeError('Must provide an ArtifactDB for artifacts to mask.')
        self.artdb = artdb

        if emldb is not None and not isinstance(emldb, EmissionLineDB):
            raise TypeError('Must provide EmissionLineDB for emission-lines to mask.')
        self.emldb = emldb

        if waverange is not None and len(waverange) != 2:
            raise ValueError('Provided wavelength range must have two and only two elements.')
        self.waverange = waverange

        ## KHRR added nsig here and above
        self.nsig = 3. if nsig is None else nsig


    def _waverange_mask(self, wave, nspec=None):
        """
        Mask the pixels **not** within the selected wavelength range.
        
        Args:
            wave (numpy.ndarray): Wavelength coordinate vector
            nspec (int): (**Optional**) Number of spectra to mask.
                Default is just one.

        Returns:
            numpy.ndarray : Boolean mask of the correct shape.
        """
        if self.waverange is None:
            return self._empty_mask(wave, ny=nspec)
        return numpy.invert(self._mask_coordinate_ranges(wave, self.waverange, ny=nspec))


    def _artifact_mask(self, wave, nspec=None):
        """
        Mask the pixels in the wavelength range(s) defined by the
        artifact database.
        
        Args:
            wave (numpy.ndarray): Wavelength coordinate vector
            nspec (int): (**Optional**) Number of spectra to mask.
                Default is just one.

        Returns:
            numpy.ndarray : Boolean mask of the correct shape.
        """
        if self.artdb is None:
            return self._empty_mask(wave, ny=nspec)
        return self._mask_coordinate_ranges(wave, self.artdb['waverange'], ny=nspec)


    def _check_eml_kin_argument(self, kin):
        """
        Check and return the correct kinematics vectors.  If a single
        number is provided, the function returns the number repeated for
        each emission line.

        Args:
            kin (float, list, numpy.ndarray):
                An input set of kinematics to used by the mask.

        Returns:
            numpy.ndarray: A 1D float array of the correct shape with
            the kinematics 

        Raises:
            ValueError:
                Raised if the length of the `kin` array is not the same
                as the number of emission lines in :attr:`emldb`.

        """
        if kin is None:
            return None

        if isinstance(kin, (int,float)):
            return numpy.full(self.emldb.size, kin, dtype=float)
        if isinstance(kin, (list, numpy.ndarray)):
            if len(kin) != self.emldb.size:
                raise ValueError('Provided vector does not have a matching length.')
            return numpy.atleast_1d(kin).astype(float)
       

    def _get_emission_line_bands(self, velocity_offsets=0.0, sigma=250.0):   #, nsigma=3.0):
        r"""
        Set the emission-line masks, using the emission-line parameters
        defined by :attr:`emldb` (see
        :class:`mangadap.par.emissionlinedb.EmissionLineDB`.

        All emission lines in the database, except for those marked with
        `action==i` are masked.
        
        The center of the masked region is constructed using
        `emldb['restwave']` and the provided velocity
        (`velocity_offset`).  The velocity of any lines marked as sky
        (`action==s`) is always set to 0.
        
        The width of the mask is set to be :math:`\pm n_\sigma \sigma`
        about this center, where both :math:`n_\sigma` and
        :math:`\sigma` are parameters (`nsigma`, `sigma`).  If `sigma` is
        None, `emldb['sig']` is used for each line.

        Args:
            velocity_offsets (float, numpy.ndarray): (**Optional**) The
                velocity offset to apply to each emission-line mask.
                Must be either a single number or one number per
                emission line.  Assumed to be 0 if set to None.
            sigma (float, numpy.ndarray): (**Optional**) Line velocity
                dispersions used to set the mask width.  Must be either
                a single number or one number per emission line.  If
                None, the dispersion provided by the emission-line
                database is used (`emldb['sig']`).
            nsigma (float, numpy.ndarray): (**Optional**) The half-width
                of the band in units of the provided velocity
                dipsersions.  Must be either a single number or one
                number per emission line.  Cannot be None.

        Returns:
            numpy.ndarray : A :math:`N_l \times 2` array with the
            wavelength limits of the emission-line bands.

        Raises:
            ValueError: Raised if `nsigma` is None, or any `nsigma` is
                not greater than 0; raised if the half-width of any mask
                is not greater than 0.
        """

        # Mask everything but the lines to ignore
        indx = numpy.invert(self.emldb['action'] == 'i')
        nbands = numpy.sum(indx)
        # No lines to mask
        if nbands == 0:
            return None

        # Get the number of standard deviations to cover with the mask
        _nsigma = self._check_eml_kin_argument(self.nsig)   # used to be (nsigma)
        if _nsigma is None:
            raise ValueError('Must provide the number of sigma to cover with the mask.')
        if numpy.any(numpy.invert(_nsigma > 0)):
            raise ValueError('Must provide a non-zero number of sigma for mask hal-fwidth.')

        # Get the mask centers
        _velocity_offsets = self._check_eml_kin_argument(velocity_offsets)
        _velocity_offsets[ self.emldb['action'] == 's' ] = 0.0
        center = self.emldb['restwave'][indx] if _velocity_offsets is None else \
                    self.emldb['restwave'][indx] \
                        * (1.0 + _velocity_offsets[indx]/astropy.constants.c.to('km/s').value)
        
        # Get the mask widths
        _sigma = self._check_eml_kin_argument(sigma)
        halfwidth = _nsigma[indx] * (self.emldb['sig'][indx] if _sigma is None else _sigma[indx])
        if numpy.any(numpy.invert(halfwidth > 0)):
            raise ValueError('All emission-line mask half-widths must be larger than 0.')

        return numpy.array([ center*(1.0-halfwidth/astropy.constants.c.to('km/s').value),
                             center*(1.0+halfwidth/astropy.constants.c.to('km/s').value) ]).T


    def _emission_line_mask(self, wave, nspec=None, velocity_offsets=0.0, sigma=250.0):   #nsigma=3.0):
        """
        Mask the pixels in the wavelength range(s) defined by the
        emission-line database that has been adjusted by a set of
        velocities and dispersions.
        
        Currently, the velocity offsets are applied to all lines for
        each spectrum, whereas sigma and nsigma are applied to all
        spectra for each line.

        Args:
            wave (numpy.ndarray): Wavelength coordinate vector
            nspec (int): (**Optional**) Number of spectra to mask.
                Default is just one.
            velocity_offsets (float, numpy.ndarray): (**Optional**) One
                or more velocity offsets to apply to the emission-line
                bands on a spectrum-by-spectrum basis. Default is to
                apply no velocity offset.
            sigma (float, numpy.ndarray): (**Optional**) One or more
                velocity dispersions to use for setting the width of the
                emission-line band on a line-by-line basis.  Default is
                a width based on a dispersion of 250 km/s.
            nsigma (float, numpy.ndarray): (**Optional**) One or more
                numbers that sets the width of the band in units of the
                provided velocity dipsersions band on a line-by-line
                basis.

        Returns:
            numpy.ndarray : Boolean mask of the correct shape.
        """
        if self.emldb is None:
            return self._empty_mask(wave, ny=nspec)


        if isinstance(velocity_offsets, (list, numpy.ndarray)) and len(velocity_offsets) > 1: 
            _velocity_offsets = numpy.asarray(velocity_offsets)
            if len(_velocity_offsets) != nspec:
                raise ValueError('Velocity offsets do not match the number of spectra.')
             
            mask = self._empty_mask(wave, ny=nspec)
            for i in range(len(_velocity_offsets)):
                waverange = self._get_emission_line_bands(velocity_offsets=_velocity_offsets[i],
                                                          sigma=sigma)  #nsigma=self.nsig)
                mask[i,:] = self._mask_coordinate_ranges(wave, waverange)
            return mask

        _velocity_offsets = velocity_offsets[0] \
                if isinstance(velocity_offsets, (list, numpy.ndarray)) \
                        and len(velocity_offsets) == 1 else velocity_offsets

        waverange = self._get_emission_line_bands(velocity_offsets=_velocity_offsets, sigma=sigma)
                                                    # nsigma=self.nsig)
        return self._mask_coordinate_ranges(wave, waverange, ny=nspec)
        

    def boolean(self, wave, nspec=None, velocity_offsets=0.0, sigma=250.0):   # nsigma=3.0):
        """
        Construct the full boolean mask that includes the desired
        wavelength range, omitting artifacts, and omitting emission
        lines.

        Args:
            wave (numpy.ndarray): Wavelength coordinate vector
            nspec (int): (**Optional**) Number of spectra to mask.
                Default is just one.
            velocity_offsets (float, numpy.ndarray): (**Optional**) One
                or more velocity offsets to apply to the emission-line
                bands on a spectrum-by-spectrum basis. Default is to
                apply no velocity offset.
            sigma (float, numpy.ndarray): (**Optional**) One or more
                velocity dispersions to use for setting the width of the
                emission-line band on a line-by-line basis.  Default is
                a width based on a dispersion of 250 km/s.
            nsigma (float, numpy.ndarray): (**Optional**) One or more
                numbers that sets the width of the band in units of the
                provided velocity dipsersions band on a line-by-line
                basis.

        Returns:
            numpy.ndarray : Boolean mask of the correct shape.
        """
        if nspec is not None and not nspec > 0:
            raise ValueError('Number of spectra must be larger than 0!')
        return self._waverange_mask(wave, nspec=nspec) | self._artifact_mask(wave, nspec=nspec) \
                | self._emission_line_mask(wave, nspec=nspec, velocity_offsets=velocity_offsets,
                                           sigma=sigma)   # nsigma=self.nsig)
        

    def bits(self, bitmask, wave, nspec=None, mask=None, velocity_offsets=0.0, sigma=250.0,
             # nsigma=3.0,
             waverange_flag='OUTSIDE_RANGE', art_flag='ARTIFACT',
             eml_flag='EML_REGION'):
        """
        Construct a bit mask that signifies pixels as outside the
        desired wavelength range, as being affected by an artifact, and
        being designated as an emission-line region.  To simply flag the
        pixels as a binary masked or unmasked, use :func:`boolean`.

        Args:
            bitmask (:class:`mangadap.util.bitmask.BitMask`): Bitmask
                used to flag pixels.  The flags waverange_flag,
                art_flag, and eml_flag *must* be defined in the provided
                instance of BitMask.
            wave (numpy.ndarray): Wavelength coordinate vector
            nspec (int): (**Optional**) Number of spectra to mask.
                Default is just one.
            mask (numpy.array): (**Optional**) Baseline mask to add
                pixels masks.  Should have data type that matches
                provided bitmask.  Default is to start with all pixels
                unmasked.
            velocity_offsets (float, numpy.ndarray): (**Optional**) One
                or more velocity offsets to apply to the emission-line
                bands on a spectrum-by-spectrum basis. Default is to
                apply no velocity offset.
            sigma (float, numpy.ndarray): (**Optional**) One or more
                velocity dispersions to use for setting the width of the
                emission-line band on a line-by-line basis.  Default is
                a width based on a dispersion of 250 km/s.
            nsigma (float, numpy.ndarray): (**Optional**) One or more
                numbers that sets the width of the band in units of the
                provided velocity dipsersions band on a line-by-line
                basis.
            waverange_flag (str): (**Optional**) Bitmask name used to
                flag a pixel as outside of the desired wavelength range.
                Default is 'OUTSIDE_RANGE'.
            art_flag (str): (**Optional**) Bitmask name used to flag a
                pixel being affected by an artifact.  Default is
                'ARTIFACT'.
            eml_flag (str): (**Optional**) Bitmask name used to flag a
                pixel being within an emission-line region.  Default is
                'EML_REGION'.

        Returns:
            numpy.ndarray : Bit mask of the correct shape.
        """

        # Check the bitmask type
        if not isinstance(bitmask, BitMask):
            raise TypeError('Must provide object of type BitMask.')
        if nspec is not None and not nspec > 0:
            raise ValueError('Number of spectra must be larger than 0!')

        # Get the wavelength range mask
        wavemask = self._waverange_mask(wave, nspec=nspec)
        # Check that the input mask has the same size
        if mask is not None and wavemask.shape != mask.shape:
            raise ValueError('Input mask does not have the correct shape.')

        # Get the artifact mask
        artmask = self._artifact_mask(wave, nspec=nspec)

        # Get the emission-line mask
        emlmask = self._emission_line_mask(wave, nspec=nspec, velocity_offsets=velocity_offsets,
                                           sigma=sigma)   # nsigma=self.nsig)

        # Construct and return the mask
        _mask = numpy.zeros(wavemask.shape, dtype=bitmask.minimum_dtype()) \
                    if mask is None else mask

        if numpy.sum(wavemask) > 0:
            _mask[wavemask] = bitmask.turn_on(_mask[wavemask], flag=waverange_flag)
        if numpy.sum(artmask) > 0:
            _mask[artmask] = bitmask.turn_on(_mask[artmask], flag=art_flag)
        if numpy.sum(emlmask) > 0:
            _mask[emlmask] = bitmask.turn_on(_mask[emlmask], flag=eml_flag)
        return _mask






from IPython import embed

from mangadap.par.artifactdb import ArtifactDB


def test_read():
    dbs = ArtifactDB.available_databases()
    assert len(dbs) > 0, 'No artifact databases available'
    for key in dbs.keys():
        artdb = ArtifactDB.from_key(key)


def test_badsky():
    artdb = ArtifactDB.from_key('BADSKY')
    assert len(artdb) == 1, 'Incorrect number of bad sky bands'
    assert 'SKY' in artdb['name'], 'Does not contain SKY in list'




r"""
Class defining the I/O configuration for MaNGA files in the DAP.

----        

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import time
import warnings
from pathlib import Path
from configparser import ConfigParser

from IPython import embed

import numpy
from astropy.io import fits

from . import defaults
from . import manga_environ
from .analysisplan import AnalysisPlan
from ..util.parser import DefaultConfig
from ..util.constants import DAPConstants
from ..util.filter import interpolate_masked_vector

def drp_version():
    """
    Return the DRP version defined by the environmental variable
    MANGADRP_VER.
    """
    return manga_environ['MANGADRP_VER']


def drp_redux_path(drpver=None):
    """
    Return the main output path for the DRP products using the
    environmental variable MANGA_SPECTRO_REDUX.

    Args:
        drpver (:obj:`str`, optional):
            DRP version. Default is to use :func:`drp_version`.

    Returns:
        :obj:`str`: Path to reduction directory
    """
    # Make sure the DRP version is set
    if drpver is None:
        drpver = drp_version()
    return Path(manga_environ['MANGA_SPECTRO_REDUX']).resolve() / drpver


def drp_directory_path(plate, drpver=None, redux_path=None):
    """
    Return the exact directory path with the DRP file.

    Args:
        plate (:obj:`int`):
            Plate number
        drpver (:obj:`str`, optional):
            DRP version. Default is to use :func:`drp_version`.
        redux_path (:obj:`str`, optional):
            Path to the root reduction directory. Default is to use
            :func:`drp_redux_path`.

    Returns:
        :obj:`str`: Path to the directory with the 3D products of the
        DRP
    """
    # Make sure the redux path is set
    _redux_path = drp_redux_path(drpver=drpver) \
                        if redux_path is None else Path(redux_path).resolve()
    return _redux_path / str(plate) / 'stack'


def drp_finding_chart_path(plate, drpver=None, redux_path=None):
    """
    Return the exact directory path with the finding charts for a given plate.

    Args:
        plate (:obj:`int`):
            Plate number
        drpver (:obj:`str`, optional):
            DRP version. Default is to use :func:`drp_version`.
        redux_path (:obj:`str`, optional):
            Path to the root reduction directory. Default is to use
            :func:`drp_redux_path`.

    Returns:
        :obj:`str`: Path to the directory with the finding charts.
    """
    # Make sure the redux path is set
    _redux_path = drp_redux_path(drpver=drpver) \
                        if redux_path is None else Path(redux_path).resolve()
    return _redux_path / str(plate) / 'images'


def drpall_file(drpver=None, redux_path=None):
    """
    Return the path to the DRPall file.

    Args:
        drpver (:obj:`str`, optional):
            DRP version.  Default is to use :func:`drp_version`.
        redux_path (:obj:`str`, optional):
            Path to the root reduction directory. Default is to use
            :func:`drp_redux_path`.

    Returns:
        :obj:`str`: Full path to the DRPall fits file.
    """
    _drpver = drp_version() if drpver is None else drpver
    _redux_path = drp_redux_path(drpver=_drpver) \
                        if redux_path is None else Path(redux_path).resolve()
    return _redux_path / f'drpall-{_drpver}.fits'


def dap_version():
    """
    Return the DAP version defined by the environmental variable
    MANGADAP_VER.  If that environmental variable does not exist,
    `mangadap.__version__` is returned.
    """
    return manga_environ['MANGADAP_VER']


def dap_analysis_path(drpver=None, dapver=None):
    """
    Return the main output path for the DAP using the environmental
    variable MANGA_SPECTRO_ANALYSIS.

    Args:
        drpver (:obj:`str`, optional):
            DRP version. Default is to use :func:`drp_version`.
        dapver (:obj:`str`, optional):
            DAP version. Default is to use :func:`dap_version`.

    Returns:
        :obj:`str`: Path to analysis directory
    """
    # Make sure the DRP version is set
    if drpver is None:
        drpver = drp_version()
    # Make sure the DAP version is set
    if dapver is None:
        dapver = dap_version()
    return Path(manga_environ['MANGA_SPECTRO_ANALYSIS']).resolve() / drpver / dapver


def manga_fits_root(plate, ifudesign, mode=None):
    """
    Generate the main root name for the output MaNGA fits files for a
    given plate/ifudesign/mode.

    Args:
        plate (:obj:`int`):
            Plate number
        ifudesign (:obj:`int`):
            IFU design number
        mode (:obj:`str`, optional):
            Mode of the output fits file. Options are: ``'LINCUBE'``,
            ``'LINRSS'``, ``'LOGCUBE'``, ``'LOGRSS'``, or ``'MAPS'``.
            Default is that no mode is included in the name.

    Returns:
        :obj:`str`: Root name for a MaNGA fits file:
        ``manga-[PLATE]-[IFUDESIGN]-[MODE]``

    Raises:
        ValueError:
            Raised if mode is not a valid option.
    """
    if mode not in [ None, 'LINCUBE', 'LINRSS', 'LOGCUBE', 'LOGRSS', 'MAPS' ]:
        raise ValueError('Do not understand mode={0}.'.format(mode))
    return f'manga-{plate}-{ifudesign}' if mode is None else f'manga-{plate}-{ifudesign}-{mode}'


def plate_target_files():
    """
    Return the default plateTarget files in mangacore and their
    associated catalog indices. The catalog indices are determined
    assuming the file names are of the form::

        'plateTargets-{0}.par'.format(catalog_id)

    Returns:
        `numpy.ndarray`_: Two arrays: the first contains the
        identified plateTargets files found using the default search
        string, the second provides the integer catalog index
        determined for each file.
    """
    core_dir = Path(manga_environ['MANGACORE_DIR']).resolve()
    if not core_dir.exists():
        raise ValueError(f'MANGACORE directory does not exist! {core_dir}')

    # Default search string
    search_str =  core_dir / 'platedesign' / 'platetargets'
    file_list = sorted(list(search_path.glob('plateTargets*.par')))
    nfiles = len(file_list)
    if nfiles == 0:
        raise ValueError('Unable to find any plateTargets files!')
    trgid = numpy.zeros(nfiles, dtype=int)                  # Array to hold indices
    for i in range(nfiles):
        suffix = file_list[i].split('-')[1]                 # Strip out the '{i}.par'
        trgid[i] = int(suffix[:suffix.find('.')])           # Strip out '{i}'
    
    return numpy.array(file_list), trgid


def redshift_fix_file():
    """
    Return the path to the default redshift fix file.

    Returns:
        :obj:`str`: Expected path to the redshift-fix parameter file.
    """
    return defaults.dap_data_root() / 'fix' / 'redshift_fix.par'


def photometry_fix_file():
    """
    Return the path to the default photometry fix file.

    Returns:
        :obj:`str`: Expected path to the photometry-fix parameter file.
    """
    return defaults.dap_data_root() / 'fix' / 'photometry_fix.par'


def parse_plate_ifu_from_file(name):
    """
    Parse the plate and ifu numbers from the file name of a MaNGA DRP file.
    """
    return tuple([int(n) for n in name.split('-')[1:3]])


# USED BY:
#   - scripts/rundap.py
def dap_config(plate, ifudesign, drpver=None, dapver=None, analysis_path=None,
               directory_path=None):
    """
    Return the full path to the DAP configuration file.

    The configuration file provides the input data necessary to
    instantiate a :class:`mangadap.datacube.manga.MaNGADataCube`.
    
    Args:
        plate (:obj:`int`):
            Plate number
        ifudesign (:obj:`int`):
            IFU design number
        drpver (:obj:`str`, optional):
            DRP version. Default is to use :func:`drp_version`.
        dapver (:obj:`str`, optional):
            DAP version. Default is to use :func:`dap_version`.
        analysis_path (:obj:`str`, optional): 
            Path to the root analysis directory. Default is to use
            :func:`dap_analysis_path`.
        directory_path (:obj:`str`, optional):
            Path to the directory with the DAP output files. Default
            is to use :func:`dap_common_path`

    Returns:
        :obj:`str`: Full path to the DAP par file
    """
    # Make sure the directory path is defined
    _directory_path = dap_common_path(plate=plate, ifudesign=ifudesign, drpver=drpver,
                                      dapver=dapver, analysis_path=analysis_path) \
                            if directory_path is None else Path(directory_path).resolve()
    # Set the name of the par file; put this in its own function?
    return _directory_path /  f'{dap_file_root(plate, ifudesign, mode="CUBE")}.ini'


# USED BY:
#   - scripts/dapall_qa.py
#   - scripts/find_repeat_observations.py
#   - survey/dapall.py
def dapall_file(drpver=None, dapver=None, analysis_path=None):
    """
    Return the path to the DAPall file.

    Args:
        drpver (:obj:`str`, optional):
            DRP version.  Default is to use :func:`drp_version`.
        dapver (:obj:`str`, optional):
            DAP version.  Default is to use :func:`dap_version`.
        analysis_path (:obj:`str`, optional):
            Path to the root analysis directory.  Default is to use
            :func:`dap_analysis_path`

    Returns:
        :obj:`str`: Full path to the DAPall fits file.
    """
    _drpver = drp_version() if drpver is None else drpver
    _dapver = dap_version() if dapver is None else dapver
    _analysis_path = dap_analysis_path(drpver=_drpver, dapver=_dapver) \
                        if analysis_path is None else Path(analysis_path).resolve()
    return _analysis_path / f'dapall-{_drpver}-{_dapver}.fits'


class MaNGAConfig:

    instrument = 'manga'

    def __init__(self, plate, ifudesign, mode='CUBE', log=True, drpver=None, redux_path=None,
                 chart_path=None, directory_path=None):
        """
        Initialize the I/O configuration using the same file as used to read
        MaNGA data.
        """
        # MaNGA-specific attributes
        self.plate = plate
        self.ifudesign = ifudesign
        MaNGAConfig.check_mode(mode)
        self.mode = mode
        self.log = log
        self.drpver = drp_version() if drpver is None else drpver
        # TODO: Depending on what is provided to the function, redux path, chart
        # path, and directory path can all be inconsistent.
        self.redux_path = drp_redux_path(drpver=self.drpver) \
                            if redux_path is None else Path(redux_path).resolve()
        self.chart_path = drp_finding_chart_path(self.plate, drpver=self.drpver,
                                                          redux_path=self.redux_path) \
                                if chart_path is None else Path(chart_path).resolve()
        self.directory_path = drp_directory_path(self.plate, drpver=self.drpver,
                                                          redux_path=self.redux_path) \
                                if directory_path is None else Path(directory_path).resolve()
        # TODO: This is only relevant the CUBE data
        if self.directory_path != drp_directory_path(self.plate, drpver=self.drpver,
                                                     redux_path=redux_path):
            warnings.warn('Paths do not match MaNGA defaults.  May not be able to find paired '
                          'CUBE & RSS files, if requested.')

    @classmethod
    def from_config(cls, cfgfile, mode='CUBE'):
        """
        """
        # Read the configuration file.  This checks if the file exists and will
        # raise an exception if it does not.
        cfg = DefaultConfig(cfgfile, interpolate=True)

        # Set the attributes, forcing a known type
        plate = cfg.getint('plate')
        ifu = cfg.getint('ifu')
        if plate is None or ifu is None:
            raise ValueError('Configuration file must define the plate and IFU.')
        log = cfg.getbool('log', default=True)
        drpver = cfg.get('drpver')
        redux_path = cfg.get('redux_path')
        directory_path = cfg.get('directory_path')

        return cls(plate, ifu, mode=mode, log=log, drpver=drpver, redux_path=redux_path,
                   directory_path=directory_path)

    @classmethod
    def from_file(cls, datafile):
        """
        Initialize the I/O configuration from the datafile for the DAP to
        process.
        """
        ifile = Path(datafile).resolve()
        if not ifile.exists():
            raise FileNotFoundError(f'File does not exist: {ifile}')

        # Parse the relevant information from the filename
        plate, ifudesign, log = ifile.name.split('-')[1:4]

        with fits.open(str(ifile)) as hdu:
            drpver = hdu[0].header['VERSDRP3']

        return cls(int(plate), int(ifudesign), mode='CUBE' if 'CUBE' in log else 'RSS',
                   log='LOG' in log, drpver=drpver, directory_path=ifile.parent)

    def copy(self):
        """
        Create a deep copy.
        """
        return MaNGAConfig(self.plate, self.ifudesign, log=self.log, drpver=self.drpver,
                           directory_path=self.directory_path)

    @property
    def file_name(self):
        root = manga_fits_root(self.plate, self.ifudesign,
                               mode=f'{"LOG" if self.log else "LIN"}{self.mode}')
        return f'{root}.fits.gz'

    @property
    def file_path(self):
        return self.directory_path / self.file_name

    @property
    def image_path(self):
        return self.chart_path / f'{self.ifudesign}.png'

    @staticmethod
    def check_mode(mode):
        """
        Check that the mode is valid.

        Valide modes are set by :func:`mode_options`.

        Args:
            mode (:obj:`str`):
                Mode value to check.

        Raises:
            ValueError:
                Raised if the mode is undefined.
        """
        options = ['CUBE', 'RSS']
        if mode not in options:
            raise ValueError(f'Unknown mode {mode}.  Must be in: {options}')

    @property
    def output_root(self):
        return f'{self.instrument}-{self.plate}-{self.ifudesign}'

    @property
    def cfg_root(self):
        """
        Generate the root name of the MaNGA DAP configuration and script files.
    
        Returns:
            :obj:`str`: Root name for the DAP file:
            ``mangadap-[PLATE]-[IFUDESIGN]-LOG[MODE]``
        """
        return f'mangadap-{self.plate}-{self.ifudesign}-LOG{self.mode}'

    @staticmethod
    def propagate_flags():
        """Flags that should be propagated from the observed data to the analysis data."""
        return ['FORESTAR']

    @staticmethod
    def do_not_use_flags():
        """Return the maskbit names that should not be used."""
        return ['DONOTUSE', 'FORESTAR']

    @staticmethod
    def do_not_fit_flags():
        """Return the maskbit names that should not be fit."""
        return ['DONOTUSE', 'FORESTAR']

    @staticmethod
    def do_not_stack_flags():
        """Return the maskbit names that should not be stacked."""
        return ['DONOTUSE', 'FORESTAR']

    @staticmethod
    def cube_pixelscale():
        """
        Return the default pixel scale of the DRP CUBE files in arcsec.
        """
        return 0.5

    @staticmethod
    def cube_width_buffer():
        """
        Return the default width buffer in pixels used when regridding the DRP
        RSS spectra into the CUBE format.
        """
        return 10

    @staticmethod
    def cube_recenter():
        """
        Return the default recentering flag used when regridding the DRP RSS
        spectra into the CUBE format.
        """
        return False

    @staticmethod
    def regrid_rlim():
        """
        Return the default limiting radius for the Gaussian kernel used when
        regridding the DRP RSS spectra into the CUBE format.
        """
        return 1.6

    @staticmethod
    def regrid_sigma():
        """
        Return the default standard deviation of the Gaussian kernel used when
        regridding the DRP RSS spectra into the CUBE format.
        """
        return 0.7

    @staticmethod
    def spectral_resolution_extension(hdu, ext=None):
        """
        Determine the spectral resolution channel to use.

        Precedence follows this order: ``LSFPRE``, ``PRESPECRES``,
        ``LSFPOST``, ``SPECRES``.

        Args:
            hdu (`astropy.io.fits.HDUList`):
                The opened MaNGA DRP file.
            ext (:obj:`str`, optional):
                Specify the extension with the spectral estimate to
                use. Should be in None, ``LSFPRE``, ``PRESPECRES``,
                ``LSFPOST``, or ``SPECRES``. The default is None,
                which means it will return the extension found first
                in the order above. None is returned if none of the
                extensions are present.

        Returns:
            :obj:`str`: The name of the preferred extension to use.
        """
        allowed = ['LSFPRE', 'LSFPOST', 'PRESPECRES', 'SPECRES']
        if ext is not None and ext not in allowed:
            warnings.warn(f'{ext} is not a viable spectral resolution extension.')
            return None
        available = [h.name for h in hdu if h.name in allowed]
        _ext = ext
        if ext is None:
            _ext = 'LSFPRE'
            if _ext not in available:
                _ext = 'PRESPECRES'
            if _ext not in available:
                _ext = 'LSFPOST'
            if _ext not in available:
                _ext = 'SPECRES'
        return _ext

    @staticmethod
    def spectral_resolution(hdu, ext=None, fill=False, median=False):
        """
        Return the spectral resolution for all spectra.

        See :func:`spectral_resolution_extension` for a description
        of the precedence used when ``ext`` is None.

        Args:
            hdu (`astropy.io.fits.HDUList`):
                The opened MaNGA DRP file.
            ext (:obj:`str`, optional):
                Specify the extension with the spectral estimate to
                use. See :func:`spectral_resolution_extension`.
            fill (:obj:`bool`, optional):
                Fill masked values by interpolation.  Default is to
                leave masked pixels in returned array.
            median (:obj:`bool`, optional):
                Return a single vector with the median spectral
                resolution instead of a per spectrum array. When using
                the ``SPECRES`` extension, this just returns the vector
                provided by the DRP file; when using either of the
                ``LSF`` extensions, this performs a masked median across
                the array and then interpolates any wavelengths that
                were masked in all vectors.

        Returns:
            :obj:`tuple`: Returns a :obj:`str` with the name of the
            extension used for the spectral-resolution measurements
            and a `numpy.ma.MaskedArray`_ with the spectral
            resolution data. Even if interpolated such that there
            should be no masked values, the function returns a masked
            array. Array contains the spectral resolution (:math:`R =
            \lambda/\Delta\lambda`) pulled from the DRP file.
        """
        # Determine which spectral resolution element to use
        _ext = MaNGAConfig.spectral_resolution_extension(hdu, ext=ext)
        # If no valid extension, raise an exception
        if ext is None and _ext is None:
            raise ValueError('No valid spectral resolution extension.')
        if ext is not None and _ext is None:
            raise ValueError('No extension: {0}'.format(ext))
        warnings.warn('Extension {0} used to define spectral resolution.'.format(_ext))
            
        # Set the mode based on the shape of the flux extension
        mode = 'CUBE' if hdu['FLUX'].data.ndim == 3 else 'RSS'

        # Build the spectral resolution vectors
        if 'SPECRES' in _ext:
            sres = numpy.ma.MaskedArray(hdu[_ext].data, mask=numpy.invert(hdu[_ext].data > 0))
            if fill:
                sres = numpy.ma.MaskedArray(interpolate_masked_vector(sres))
            if not median:
                sres = numpy.tile(sres, (hdu['FLUX'].data.shape[0],1)) if mode == 'RSS' \
                         else numpy.tile(sres, (*hdu['FLUX'].data.shape[1:][::-1],1)).T
            return _ext, sres

        # Otherwise dealing with the DISP data
        sres = numpy.ma.MaskedArray(hdu[_ext].data)
        # Mask any non-positive value
        sres[numpy.invert(sres > 0)] = numpy.ma.masked
        # Convert from sigma in angstroms to spectral resolution (based
        # on FWHM). Make sure to treat array shapes correctly for the
        # different modes.
        if mode == 'RSS':
            sres = numpy.ma.divide(hdu['WAVE'].data[None,:], sres) / DAPConstants.sig2fwhm
        else:
            # Mode is 'CUBE'
            sres = numpy.ma.divide(hdu['WAVE'].data[:,None,None], sres) / DAPConstants.sig2fwhm
        # Interpolate over any masked values
        if fill:
            axis = 1 if mode == 'RSS' else 0
            sres = numpy.ma.MaskedArray(
                        numpy.ma.apply_along_axis(interpolate_masked_vector, axis, sres))
        if median:
            sres = numpy.ma.median(sres, axis=0) if mode == 'RSS' \
                        else numpy.ma.median(sres.reshape(sres.shape[0],-1), axis=1)
        return _ext, sres

    @staticmethod
    def write_config(ofile, plate, ifudesign, log=True, z=None, vdisp=None, ell=None, pa=None,
                     reff=None, sres_ext=None, sres_fill=None, covar_ext=None, drpver=None,
                     redux_path=None, directory_path=None, overwrite=True):
        """
        Write the configuration file that can be used to instantiate
        a MaNGA data object
        (:class:`mangadap.datacube.manga.MaNGADataCube` or
        :class:`mangadap.spectra.manga.MaNGARSS`).
        
        See :func:`from_config`.

        Args:
            ofile (:obj:`str`, optional):
                Name of the configuration file.
            plate (:obj:`int`):
                Plate number
            ifudesign (:obj:`int`):
                IFU design
            log (:obj:`bool`, optional):
                Use the datacube that is logarithmically binned in
                wavelength.
            z (:obj:`float`, optional):
                Estimated bulk redshift. If None, some of the DAP
                analysis modules will fault.
            vdisp (:obj:`float`, optional):
                Estimated velocity dispersion. If None, some of the
                DAP analysis modules will assume an initial guess of
                100 km/s.
            ell (:obj:`float`, optional):
                Characteristic isophotal ellipticity (1-b/a). If
                None, some of the DAP modules will issue a warning
                and continue by assuming ``ell=0``.
            pa (:obj:`float`, optional):
                Characteristic isophotal position angle (through E
                from N). If None, some of the DAP modules will issue
                a warning and continue by assuming ``pa=0``.
            reff (:obj:`float`, optional):
                Effective (half-light) radius in arcsec. If None,
                some of the DAP modules will issue a warning and
                continue by assuming ``reff=1``.
            sres_ext (:obj:`str`, optional):
                The extension to use when constructing the spectral
                resolution vectors. See :func:`spectral_resolution`.
            sres_fill (:obj:`bool`, optional):
                Fill masked values by interpolation. Default is to
                leave masked pixels in returned array.
            covar_ext (:obj:`str`, optional):
                Extension to use as the single spatial correlation
                matrix for all wavelength channels, read from the DRP
                file. For generating the covariance matrix directly
                for an arbitrary wavelength channel using the RSS
                file, see
                :func:`mangadap.datacube.datacube.DataCube.covariance_matrix`.
            drpver (:obj:`str`, optional):
                DRP version, which is used to define the default DRP
                redux path. Default is defined by
                :func:`mangadap.config.manga.drp_version`
            redux_path (:obj:`str`, optional):
                The path to the top level directory containing the
                DRP output files for a given DRP version. Default is
                defined by
                :func:`mangadap.config.manga.drp_redux_path`.
            directory_path (:obj:`str`, optional):
                The exact path to the DRP file. Default is defined by
                :func:`mangadap.config.manga.drp_directory_path`.
                Providing this ignores anything provided for
                ``drpver`` or ``redux_path``.
            overwrite (:obj:`bool`, optional):
                Overwrite any existing parameter file.
        """
        _ofile = Path(ofile).resolve()
        if _ofile.exists() and not overwrite:
            raise FileExistsError(f'{_ofile} already exists; to overwrite, set '
                                  'overwrite=True.')

        if not _ofile.parent.exists():
            _ofile.parent.mkdir(parents=True)

        # Build the configuration data
        cfg = ConfigParser(allow_no_value=True)
        cfg['default'] = {'drpver': drpver,
                          'redux_path': redux_path,
                          'directory_path': directory_path,
                          'plate': str(plate),
                          'ifu': str(ifudesign),
                          'log': str(log),
                          'sres_ext': sres_ext,
                          'sres_fill': sres_fill,
                          'covar_ext': covar_ext,
                          'z': None if z is None else '{0:.7e}'.format(z),
                          'vdisp': None if vdisp is None else '{0:.7e}'.format(vdisp),
                          'ell': None if ell is None else '{0:.7e}'.format(ell),
                          'pa': None if pa is None else '{0:.7e}'.format(pa),
                          'reff': None if reff is None else '{0:.7e}'.format(reff)}

        # Write the configuration file
        with _ofile.open('w') as f:
            f.write('# Auto-generated configuration file\n')
            f.write('# {0}\n'.format(time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())))
            f.write('\n')
            cfg.write(f)


class MaNGAAnalysisPlan(AnalysisPlan):
    """
    Define the analysis plan and output paths for MaNGA data.

    This class only redefines the output paths from the base class.
    """
    def __init__(self, planlist, cube=None, analysis_path=None):
#        if cube is None:
#            raise ValueError('For a MaNGA analysis plan, you must provide the cube because it '
#                             'is used to define the output paths.')
        self.cube = cube
        if self.cube is None:
            warnings.warn('Instantiation of the MaNGA analysis plan did not include the cube.  '
                          'Directories will be *undefined*, meaning the code may fault.')
        if analysis_path is None:
            analysis_path = dap_analysis_path(drpver=cube.drpver)
        super().__init__(planlist, cube=cube, analysis_path=analysis_path)

    def common_path(self):
        """
        Return the path for data common to all plans.

        Args:
            cube (:class:`~mangadap.datacube.datacube.DataCube`, optional):
                Cube being analyzed.  Passed for cube-specific path
                specification.  Not used by this base class.

        Returns:
            `Path`_: Path object for the "common" output
        """
        if self.cube is None:
            raise ValueError('Path undefined because cube being processed was not provided '
                             'when MaNGAAnalysisPlan was instantiated!')
        return self.analysis_path / 'common' / str(self.cube.plate) / str(self.cube.ifudesign)

    def method_path(self, plan_index=0, qa=False, ref=False):
        """
        Return the path for method-specific output.

        Args:
            cube (:class:`~mangadap.datacube.datacube.DataCube`, optional):
                Cube being analyzed.  Passed for cube-specific path
                specification.  Not used by this base class.
            plan_index (:obj:`int`, optional):
                The index of the plan.  This is used to select the 'key' of the
                analysis plan being used, which is used as the subdirectory for
                the output.
            qa (:obj:`bool`, optional):
                Flag that the output is specifically quality assessment plots.
            ref (:obj:`bool`, optional):
                Flag that the output is specifically reference data.

        Returns:
            `Path`_: Path object for the method-specific output

        Raises:
            ValueError:
                Raised if the plan index is invalid or if both qa and ref are true.

        """
        if self.cube is None:
            raise ValueError('Path undefined because cube being processed was not provided '
                             'when MaNGAAnalysisPlan was instantiated!')
        if plan_index < 0 or plan_index >= self.nplans:
            raise ValueError(f'Invalid index ({plan_index}); 0 <= index < {self.nplans}.')
        if qa and ref:
            raise ValueError('Cannot provide path for both qa and ref directory.  Pick one.')

        root = self.analysis_path / self.plan[self.plan_keys[plan_index]]['key'] \
                    / str(self.cube.plate) / str(self.cube.ifudesign)
        if not qa and not ref:
            return root
        if qa:
            return root / 'qa'
        return root / 'ref'




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Container class for a database with a list of spectral regions with
known artifacts that should be ignored during analysis of the data.
These can be anything, but is only currently used to define spectral
regions with poorly subtracted sky lines.  They are also currently
independent of spatial position and expected to be applied for all
spectra in an RSS or CUBE file.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import os
import numpy

from pydl.goddard.astro import airtovac
from pydl.pydlutils.yanny import yanny
from .parset import KeywordParSet, ParDatabase
from .spectralfeaturedb import SpectralFeatureDB

# Add strict versioning
# from distutils.version import StrictVersion

class ArtifactPar(KeywordParSet):
    """
    Parameter object that defines a set of artifacts to be ignored
    during analysis.

    The defined parameters are:

    .. include:: ../tables/artifactpar.rst
    
    See :class:`mangadap.par.parset.ParSet` for attributes and raised
    exceptions.
    """
    def __init__(self, index=None, name=None, waverange=None):
        
        arr_like = [ numpy.ndarray, list ]

        _name = None if name is None else name.strip()

        pars =     [ 'index', 'name', 'waverange' ]
        values =   [   index,  _name,   waverange ]
        dtypes =   [     int,    str,    arr_like ]

        descr = ['An index used to refer to the artifact',
                 'A name for the artifact',
                 'A two-element vector with the starting and ending wavelength (angstroms ' \
                 'in **vacuum**) where the artifact affects the data.']

        super(ArtifactPar, self).__init__(pars, values=values, dtypes=dtypes, descr=descr)

    def _check(self):
        """
        Check the parameter list.

        Only check performed is to make sure the wavelength range
        only has two elements.

        Raises:
            ValueError:
                Raised if one of the conditions above are not met.
        """
        if len(self.data['waverange']) != 2:
            raise ValueError('Wavelength range must have two and only two elements.')


class ArtifactDB(SpectralFeatureDB):
    """
    Basic container class for the database of artifacts.

    Each row of the database is parsed using :class:`ArtifactPar`.

    The primary instantiation requires the SDSS parameter file with
    the artifact data. To instantiate using a keyword (and
    optionally a directory that holds the parameter files), use the
    :func:`mangadap.par.spectralfeaturedb.SpectralFeatureDB.from_key`
    class method.  See the base class for additional attributes.

    Args:
        parfile (:obj:`str`):
            The SDSS parameter file with the artifact database.

    Attributes:
        key (:obj:`str`):
            Database signifying keyword
        file (:obj:`str`):
            File with the artifact data
        size (:obj:`int`):
            Number of artifacts in the database. 
    """
    default_data_dir = 'artifacts'
    def _parse_yanny(self):
        """
        Parse the yanny file (provided by :attr:`file`) for the artifact
        database.

        Returns:
            :obj:`list`: The list of
            :class:`mangadap.par.parset.ParSet` instances for each
            line of the database.
        """
        # Read the yanny file
        par = yanny(filename=self.file, raw=True)
        if len(par['DAPART']['index']) == 0:
            raise ValueError('Could not find DAPART entries in {0}!'.format(self.file))

        # Setup the array of emission line database parameters
        self.size = len(par['DAPART']['index'])
        parlist = []
        for i in range(self.size):
            invac = par['DAPART']['waveref'][i] == 'vac'
            parlist += [ ArtifactPar(index=par['DAPART']['index'][i], name=par['DAPART']['name'][i],
                                     waverange=numpy.asarray(par['DAPART']['waverange'][i]) \
                                      if invac else airtovac(par['DAPEML']['waverange'][i]) )]
        return parlist



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Adapted from pypeit

# NOTE: The configuration for the package, including the name, version, and
# other information are set in the setup.cfg file.

# First provide helpful messages if contributors try and run legacy commands
# for tests or docs.

TEST_HELP = """
Note: running tests via 'python setup.py test' is now deprecated. The recommended method
is to run:

    tox -e test-alldeps

The Python version can also be specified, e.g.:

    tox -e py38-test-alldeps

You can list all available environments by doing:

    tox -a

If you don't already have tox installed, you can install it by doing:

    pip install tox

If you want to run all or part of the test suite within an existing environment,
you can use pytest directly:

    pip install -e .[dev]
    pytest

For more information, see:

  http://docs.astropy.org/en/latest/development/testguide.html#running-tests
"""

import sys
if 'test' in sys.argv:
    print(TEST_HELP)
    sys.exit(1)

VERSION_TEMPLATE = """
# Note that we need to fall back to the hard-coded version if either
# setuptools_scm can't be imported or setuptools_scm can't determine the
# version, so we catch the generic 'Exception'.
try:
    from setuptools_scm import get_version
    version = get_version(root='..', relative_to=__file__)
except Exception:
    version = '{version}'
""".lstrip()

import os
from setuptools import setup
setup(use_scm_version={'write_to': os.path.join('mangadap', 'version.py'),
                       'write_to_template': VERSION_TEMPLATE})





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
A set of functions used to filter arrays.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import warnings
from IPython import embed
import numpy
from scipy import interpolate, sparse

# Debug
from matplotlib import pyplot
from matplotlib.ticker import NullFormatter

def high_pass_filter(flux, dw=0, k=None, Dw=None):
    """
    Pulled from FIREFLY's hpf() function and edited on 17 Nov 2016.

    Apply a high-pass filter to the input vector.

    """

    n = flux.size
    _dw = int(dw)

    if k is None and Dw is None:
        Dw = 100
        _k = n//Dw
    elif k is not None and Dw is None:
        _k = int(k)
        Dw = n//_k
    elif k is None and Dw is not None:
        _k = n//Dw
    else:
        raise ValueError('Cannot provide both k and Dw.')

    # PRINT A REPORT

    # Rita's typical inputs for SDSS:
    # w = 10 # 10
    # windowsize = 20 # 20

    # My MaNGA inputs:
    # if w == 0 and windowsize == 0:
    #     print "HPF parameters not specified: choosing default window (stepsize=40)"
    #     w = 40
    #     windowsize = 0

    h           = numpy.fft.fft(flux)
    h_filtered  = numpy.zeros(n, dtype=complex)
    window      = numpy.zeros(n)
    unwindow    = numpy.zeros(n)
    window[0]   = 1                             # keep F0 for normalisation
    unwindow[0] = 1

    for i in range(_dw):
        window[_k+i] = (i+1.0)/_dw
        window[n-1-(_k+_dw-i)] = (_dw-i)/_dw
    window[_k+_dw:n-(_k+_dw)] = 1

    unwindow        = 1 - window
    unwindow[0]     = 1
    
    h_filtered      = h * window
    un_h_filtered   = h * unwindow

    res     = numpy.real(numpy.fft.ifft(h_filtered))
    unres   = numpy.real(numpy.fft.ifft(un_h_filtered)) 
    res_out = (1.0+(res-numpy.median(res))/unres) * numpy.median(res) 

    return res_out, window, res, unres

#def off_diagonal_identity(size, win):
#    r"""
#    Construct a matrix with ones within a window along the diagonal.
#
#    Args:
#        size (int) : Size for the square matrix; i.e., :math:`N` for the
#            :math:`N\timesN` matrix.
#        win (int): Number of ones in each row along the diagonal.
#
#    Raises:
#        ValueError: Raised if the window is larger than 2*size-1.
#
#    """
#    if win > 2*size-1:
#        raise ValueError('Window too large for matrix size.')
#    if win == 2*size-1:
#        return numpy.ones((size,size), dtype=int)
#    x = numpy.zeros((size,size), dtype=int)#numpy.identity(size).astype(int)
#    for i in range(1,(win+1)//2):
#        x[:-i,i:] = x[:-i,i:] + numpy.identity(size-i).astype(int)
#    x += x.T
#    if win % 2 != 1:
#        x[:-i-1,i+1:] = x[:-i-1,i+1:] + numpy.identity(size-i-1).astype(int)
#    return x + numpy.identity(size).astype(int)

def off_diagonal_identity(size, win, return_sparse=False):
    r"""
    Construct a matrix with ones within a window along the diagonal.

    Args:
        size (int) : Size for the square matrix; i.e., :math:`N` for the
            :math:`N\timesN` matrix.
        win (int): Number of ones in each row along the diagonal.

    Raises:
        ValueError: Raised if the window is larger than 2*size-1.

    """
    if win > 2*size-1:
        raise ValueError('Window too large for matrix size.')
    if win == 2*size-1:
        return numpy.ones((size,size), dtype=int)

    # Indices of diagonal
    ii = numpy.arange(size).astype(int)

    # Build the upper triangle
    i = numpy.empty(0, dtype=int)
    j = numpy.empty(0, dtype=int)
    for k in range(1,(win+1)//2):
        i = numpy.append(i, ii[:size-k])
        j = numpy.append(j, ii[k:size])

    # Copy to the lower triangle
    _i = numpy.append(i,j)
    j = numpy.append(j,i)

    # Add the diagonal
    i = numpy.append(_i, ii)
    j = numpy.append(j, ii)

    # Accommodate an even window
    if win % 2 != 1:
        i = numpy.append(i, ii[:size-k-1])
        j = numpy.append(j, ii[k+1:size])

    # Construct and return the array
    if return_sparse:
        return sparse.coo_matrix((numpy.ones(len(i), dtype=int),(i,j)), shape=(size,size)).tocsr()

    a = numpy.zeros((size,size), dtype=int)
    a[i,j] = 1
    return a


def build_smoothing_mask(x, pix_buffer, default=None, mask_x=None):
    r"""
    Construct a mask for the provided vector that masks the first and
    last pix_buffer pixels in the coordinate vector.
    
    The mask is instantiated as fully unmasked, unless a default mask is
    provided.
    
    To mask specified ranges in x, provide mask_x with a shape
    :math:`N_{\rm mask}\times 2` where each mask is defined by the
    starting and ending value of x to exclude.
    """
    # Smooth the ratio of the data to the model
    if len(x.shape) != 1:
        raise ValueError('Input must be a vector.')
    npix = x.size
    mask = numpy.zeros(npix, dtype=bool) if default is None else default.copy()
    if mask.size != npix:
        raise ValueError('Provided default has an incorrect length.')
    mask[:pix_buffer] = True
    mask[-pix_buffer:] = True
    if mask_x is None:
        return mask

    for m in mask_x:
        mask |= numpy.logical_and(x > m[0], x < m[1])
    return mask


class BoxcarFilter():
    """
    """
    def __init__(self, boxcar, lo=None, hi=None, niter=None, y=None, wgt=None, mask=None,
                 local_sigma=None):

        if boxcar <= 1:
            raise ValueError('Boxcar must be greater than 1!')

        self.y = None
        self.wgt = None
        self.input_mask = None
        self.output_mask = None
        self.boxcar = boxcar
        self.lo_rej = lo
        self.hi_rej = hi
        self.nrej = -1 if niter is None else niter
        self.nvec = None
        self.npix = None
        self.local_sigma = False if local_sigma is None else local_sigma

        self.smoothed_wgt = None
        self.smoothed_n = None
        self.smoothed_y = None
        self.sigma_y = None
        self.rdx_index = None

        if y is not None:
            self.smooth(y, wgt=wgt, mask=mask, boxcar=boxcar, lo=lo, hi=hi, niter=niter,
                        local_sigma=local_sigma)


    def _assign_par(self, boxcar, lo, hi, niter, local_sigma):
        if boxcar is not None:
            self.boxcar = boxcar
        if lo is not None:
            self.lo_rej = lo
        if hi is not None:
            self.hi_rej = hi
        if niter is not None:
            self.nrej = niter
        if local_sigma is not None:
            self.local_sigma = local_sigma


    def _check_par(self):
        if self.boxcar <= 1:
            raise ValueError('Boxcar must be greater than 1!')

        if self.nrej is not None and self.nrej > 0 and self.lo_rej is None and self.hi_rej is None:
            raise ValueError('Must provide lo or hi if niter provided')


    def _reduce_indices(self):
        indx = numpy.array([numpy.arange(self.npix)-self.boxcar//2,
                            numpy.arange(self.npix)+self.boxcar//2+1])
        if self.boxcar % 2 != 1:
            indx[:self.npix] += 1
        return indx.T.ravel().clip(0,self.npix-1)


    def _apply(self):
        self.smoothed_n = numpy.add.reduceat(numpy.invert(self.output_mask), self.rdx_index, axis=1,
                                             dtype=int)[:,::2]
        self.smoothed_wgt = numpy.add.reduceat(self.input_wgt, self.rdx_index, axis=1,
                                             dtype=float)[:,::2]
        self.smoothed_y = numpy.add.reduceat(self.input_wgt*self.y.filled(0.0), self.rdx_index,
                                             axis=1, dtype=float)[:,::2]
        smoothed_y2 = numpy.add.reduceat(self.input_wgt*numpy.square(self.y.filled(0.0)),
                                         self.rdx_index, axis=1, dtype=float)[:,::2]

        self.smoothed_y = numpy.ma.divide(self.smoothed_y, self.smoothed_wgt)
        self.smoothed_y[self.output_mask | (self.smoothed_n == 0)] = numpy.ma.masked

        self.sigma_y = numpy.ma.sqrt((numpy.ma.divide(smoothed_y2, self.smoothed_wgt)
                                            - numpy.square(self.smoothed_y))
                                        * numpy.ma.divide(self.smoothed_n, self.smoothed_n-1)) \
                        if self.local_sigma else \
                            numpy.ma.MaskedArray(numpy.array([numpy.std(self.y - self.smoothed_y,
                                                                        axis=1)]*self.npix).T)

        self.output_mask = numpy.ma.getmaskarray(self.smoothed_y) \
                                | numpy.ma.getmaskarray(self.sigma_y) 
        
        self.smoothed_y[self.output_mask | (self.smoothed_n == 0)] = numpy.ma.masked
        self.sigma_y[self.output_mask | (self.smoothed_n == 0)] = numpy.ma.masked

        return

        w,h = pyplot.figaspect(1)
        fig = pyplot.figure(figsize=(1.5*w,1.5*h))

        minf = numpy.amin( numpy.ma.log10(numpy.append(self.y.compressed(), self.smoothed_y.compressed())) )
        maxf = numpy.amax( numpy.ma.log10(numpy.append(self.y.compressed(), self.smoothed_y.compressed())) )

        mins = numpy.amin(numpy.ma.log10(self.sigma_y).compressed())
        maxs = numpy.amax(numpy.ma.log10(self.sigma_y).compressed())

        minr = numpy.amin(numpy.ma.log10(numpy.ma.divide(self.y, self.smoothed_y)).compressed())
        maxr = numpy.amax(numpy.ma.log10(numpy.ma.divide(self.y, self.smoothed_y)).compressed())

        ax = fig.add_axes([0.05, 0.50, 0.45, 0.45])
        ax.xaxis.set_major_formatter(NullFormatter())
        ax.imshow(numpy.ma.log10(self.y), origin='lower', interpolation='nearest', vmin=minf,
                  vmax=maxf, aspect='auto')
        ax = fig.add_axes([0.50, 0.50, 0.45, 0.45])
        ax.xaxis.set_major_formatter(NullFormatter())
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.imshow(numpy.ma.log10(self.smoothed_y), origin='lower', interpolation='nearest',
                  vmin=minf, vmax=maxf, aspect='auto')
        ax = fig.add_axes([0.05, 0.05, 0.45, 0.45])
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.imshow(numpy.ma.log10(numpy.ma.divide(self.y,self.smoothed_y)), origin='lower',
                  interpolation='nearest', aspect='auto', vmin=minr, vmax=maxr)
        ax = fig.add_axes([0.50, 0.05, 0.45, 0.45])
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.imshow(numpy.ma.log10(self.sigma_y), origin='lower', interpolation='nearest',
                  aspect='auto', vmin=mins, vmax=maxs)

        pyplot.show()
        exit()

    
    def _interpolate(self):
        """
        Interpolate the smoothed image across the masked regions.
        Leading and trailing masked regions of each row are set to the
        first and last unmasked value, respectively.

        interpolate both the smoothed data and the sigma

        """

        # TODO: Put in a check here in case the full array is masked...

        # Boolean arrays with bad pixels (should be the same for both
        # smoothed_y and sigma_y)
        badpix = numpy.ma.getmaskarray(self.smoothed_y)

        # Deal with any vectors that are fully masked
        goodvec = numpy.sum(numpy.invert(badpix), axis=1) > 0
        ngood = numpy.sum(goodvec)
        veci = numpy.arange(self.nvec)[goodvec]

        # Find the first and last unflagged pixels in the good vectors
        pixcoo = numpy.ma.MaskedArray(numpy.array([numpy.arange(self.npix)]*ngood),
                                      mask=badpix[goodvec,:]).astype(int)
        mini = numpy.ma.amin(pixcoo, axis=1)
        maxi = numpy.ma.amax(pixcoo, axis=1)

        # Copy those to the first and last pixels for each row
        self.smoothed_y[veci,0] = numpy.array([self.smoothed_y[i,m] for i,m in zip(veci, mini)])
        self.smoothed_y[veci,-1] = numpy.array([self.smoothed_y[i,m] for i,m in zip(veci, maxi)])

        self.sigma_y[veci,0] = numpy.array([self.sigma_y[i,m] for i,m in zip(veci, mini)])
        self.sigma_y[veci,-1] = numpy.array([self.sigma_y[i,m] for i,m in zip(veci, maxi)])

        # Detach badpix from self.smoothed_y
        badpix = numpy.ma.getmaskarray(self.smoothed_y).copy()

        # Interpolate the smoothed array
        x = numpy.ma.MaskedArray(numpy.arange(self.nvec*self.npix), mask=badpix)
        interpolator = interpolate.interp1d(x.compressed(), self.smoothed_y.compressed(),
                                            assume_sorted=True, fill_value='extrapolate')
        self.smoothed_y.ravel()[badpix.ravel()] = interpolator(x.data[badpix.ravel()])
        self.smoothed_y[numpy.invert(goodvec),:] = 0.0

        # Interpolate the sigma array
        interpolator = interpolate.interp1d(x.compressed(), self.sigma_y.compressed(),
                                            assume_sorted=True, fill_value='extrapolate')
        self.sigma_y.ravel()[badpix.ravel()] = interpolator(x.data[badpix.ravel()])
        self.sigma_y[numpy.invert(goodvec),:] = 0.0


    def smooth(self, y, wgt=None, mask=None, boxcar=None, lo=None, hi=None, niter=None,
               local_sigma=None):
        """
        Smooth a vector or array of vectors by a boxcar with specified
        rejection.
        """
        # Assign and check the binning parameters
        self._assign_par(boxcar, lo, hi, niter, local_sigma)
        self._check_par()

        # Check the input vector/array
        if len(y.shape) > 2:
            raise ValueError('Can only deal with vectors or matrices.')

        # Set the weighting
        self.input_wgt = numpy.ones(y.shape, dtype=bool) if wgt is None else wgt
        if self.input_wgt.shape != y.shape:
            raise ValueError('Weights shape does not match data.')
        if isinstance(y, numpy.ma.MaskedArray):
            self.input_wgt[numpy.ma.getmaskarray(y)] = 0.0

        # Set the mask
        self.input_mask = numpy.zeros(y.shape, dtype=bool) if mask is None else mask
        if self.input_mask.shape != y.shape:
            raise ValueError('Mask shape does not match data.')
        if isinstance(y, numpy.ma.MaskedArray):
            self.input_mask |= numpy.ma.getmaskarray(y)

        # Save the input
        provided_vector = len(y.shape) == 1
        self.y = numpy.ma.atleast_2d(numpy.ma.MaskedArray(y.copy(), mask=self.input_mask))
        self.input_wgt = numpy.atleast_2d(self.input_wgt)
        self.input_mask = numpy.atleast_2d(self.input_mask)
        self.output_mask = self.input_mask.copy()
        self.nvec, self.npix = self.y.shape

        # Create the list of reduceat indices
        self.rdx_index = self._reduce_indices()

        # Get the first iteration of the smoothed vectors
        self._apply()

        # No rejection iterations requested so return the result
        if self.lo_rej is None and self.hi_rej is None:
            self._interpolate()
            return self.smoothed_y[0,:] if provided_vector else self.smoothed_y

        # Iteratively reject outliers
        nmasked = numpy.sum(numpy.ma.getmaskarray(self.y))
        i=0
        while i > -1:
            nbad = numpy.sum(self.output_mask)
            if self.lo_rej is not None:
                self.output_mask = numpy.logical_or(self.output_mask, self.y.data - self.smoothed_y
                                                        < (-self.lo_rej*self.sigma_y))
            if self.hi_rej is not None:
                self.output_mask = numpy.logical_or(self.output_mask, self.y.data - self.smoothed_y
                                                        > (self.hi_rej*self.sigma_y))
            self.y[self.output_mask] = numpy.ma.masked
            self._apply()

            nmasked = numpy.sum(self.output_mask) - nbad
#            print('Niter: {0}; Nmasked: {1}'.format(i+1, nmasked))

            i += 1
            if i == self.nrej or nmasked == 0:
                break

        self._interpolate()
        return self.smoothed_y[0,:] if provided_vector else self.smoothed_y


    @property
    def mask(self):
        return self.output_mask.copy()



def smooth_masked_vector(x, nsmooth):
    """
    Smooth a masked vector by a box of size nsmooth.
    """
    n = off_diagonal_identity(x.size, nsmooth)*numpy.invert(numpy.ma.getmaskarray(x))[None,:]
    nn = numpy.sum(n,axis=1)
    return numpy.ma.MaskedArray(numpy.dot(n,x.data), mask=numpy.invert(nn>0) | x.mask)/nn


def interpolate_masked_vector(y, quiet=True, extrap_with_median=False):
    """
    Interpolate over the masked pixels in an input vector using linear
    interpolation.

    Args:
        y (numpy.ma.MaskedArray):
            Vector to interpolate.
        quiet (:obj:`bool`, optional):
            Suppress standard output (default is True).
        extrap_with_median (:obj:`bool`, optional):
            Instead of linear extrapolation, use the median of the
            interpolated vector.  Defaults to linear extrapolation.

    Returns:
        numpy.ndarray: Returns the interpolated, unmasked vector.

    """
    x = numpy.arange(y.size)
    indx = numpy.ma.getmaskarray(y)
    if numpy.sum(numpy.invert(indx)) < 2:
        if not quiet:
            warnings.warn('Input vector has fewer than 2 unmasked values!  Returning zero vector.')
        return numpy.zeros(y.size, dtype=y.dtype.name)
    interpolator = interpolate.interp1d(x[numpy.invert(indx)], y[numpy.invert(indx)],
                                        fill_value='extrapolate')
    _y = y.data.copy()
    _y[indx] = interpolator(x[indx])

    if extrap_with_median:
        med = numpy.median(interpolator.y)
        m_indx = (x[indx] < interpolator.x[0]) | (x[indx] > interpolator.x[-1])
        _y[indx][m_indx] = med

    return _y


def boxcar_smooth_vector(x, boxcar, mask=None, lo=None, hi=None, niter=None, return_mask=False):
    """
    Boxcar smooth an input vector.
        - Ignores masked pixels.
        - Allows for iterative positive and negative outliers.
        - Can return the mask separately from the smoothed vector.
    """
    if niter is not None and lo is None and hi is None:
        raise ValueError('Must provide lo or hi if niter provided')

    _mask = numpy.zeros(x.size, dtype=bool) if mask is None else mask
    _x = x if isinstance(x, numpy.ma.MaskedArray) else numpy.ma.MaskedArray(x)
    _x[_mask] = numpy.ma.masked

    sx = interpolate_masked_vector( smooth_masked_vector(_x, boxcar) )

#    pyplot.step(numpy.arange(_x.size), _x.data, where='mid', color='0.3', lw=0.5)
#    pyplot.step(numpy.arange(_x.size), _x, where='mid', color='k', lw=0.5)
#    pyplot.plot(numpy.arange(_x.size), sx, color='r', lw=1)
#    pyplot.show()

    if numpy.all([ x is None for x in [ lo, hi, niter ]]):
        if return_mask:
            return sx, numpy.ma.getmaskarray(_x).copy()
        return sx
    
    _niter = -1 if niter is None else niter
    nrej = numpy.sum(numpy.ma.getmaskarray(_x))
    i=0
    while i > -1:
        sig = numpy.std((_x - sx).compressed())
        mask = numpy.ma.getmaskarray(_x).copy()
        if lo is not None:
            mask = numpy.logical_or(mask, _x.data - sx < -lo*sig)
        if hi is not None:
            mask = numpy.logical_or(mask, _x.data - sx > hi*sig)
        nrej = numpy.sum(mask) - numpy.sum(_x.mask)
#        print('Niter: {0}; Nrej: {1}'.format(i+1, nrej))
        _x[mask] = numpy.ma.masked
        sx = interpolate_masked_vector( smooth_masked_vector(_x, boxcar) )

#        pyplot.step(numpy.arange(_x.size), _x.data, where='mid', color='cyan', lw=0.5)
#        pyplot.step(numpy.arange(_x.size), _x, where='mid', color='k', lw=0.5)
#        pyplot.plot(numpy.arange(_x.size), sx, color='r', lw=1)
#        pyplot.show()

        i += 1
        if i == _niter or nrej == 0:
            break

    pyplot.step(numpy.arange(_x.size), _x.data, where='mid', color='cyan', lw=0.5)
    pyplot.step(numpy.arange(_x.size), _x, where='mid', color='k', lw=0.5)
    pyplot.plot(numpy.arange(_x.size), sx, color='r', lw=1)
    pyplot.show()

    if return_mask:
        return sx, numpy.ma.getmaskarray(_x).copy()
    return sx





import time
import warnings

from IPython import embed

import numpy
from scipy import interpolate
from matplotlib import pyplot, colors, rc, colorbar, ticker, cm

from astropy.io import fits

from mangadap.datacube import DataCube
from mangadap.config.analysisplan import AnalysisPlan
from mangadap.proc.templatelibrary import TemplateLibraryDef
from mangadap.proc.reductionassessments import ReductionAssessment, ReductionAssessmentDef
from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectraDef
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel, StellarContinuumModelDef
from mangadap.proc.util import growth_lim
from mangadap.util.fitsutil import DAPFitsUtil
from mangadap.util.mapping import map_extent, map_beam_patch
from mangadap.util.pkg import load_object
from mangadap.config import defaults
from mangadap import dapfits

from mangadap.scripts import scriptbase


def init_image_ax(fig, pos):
    ax = fig.add_axes(pos, facecolor='0.9')
    ax.minorticks_on()
    ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6, direction='in')
    ax.tick_params(which='minor', length=3, direction='in')
    return ax


# THIS IS A TEST FUNCTION!
def stellar_continuum_scatter(plt, ifu, daptype, binid, snr, r68, r99, rchi2,
                              signal, a, da, an, dan, gmr, t, dt, tn, dtn,
                              svel, ssigo, ssigcor, ssigc, extent=None, ofile=None):

    snr_lim = numpy.power(10., growth_lim(numpy.ma.log10(snr), 1.00, fac=1.05))
    snr_lim[0] = max(0.1, snr_lim[0])
    r68_lim = numpy.power(10., growth_lim(numpy.ma.log10(r68), 0.90, fac=1.05))
    r99_lim = numpy.power(10., growth_lim(numpy.ma.log10(r99), 0.90, fac=1.05))
    chi_lim = growth_lim(rchi2, 0.90, fac=1.05)

    s_lim = numpy.power(10., growth_lim(numpy.ma.log10(signal), 1.00, fac=1.10))
    a_lim = growth_lim(numpy.ma.log10(a), 0.99, fac=1.10)
    da_lim = growth_lim(numpy.ma.log10(da), 0.99, fac=1.10)
    an_lim = growth_lim(numpy.ma.log10(an), 1.00, fac=1.10)
    dan_lim = growth_lim(numpy.ma.log10(dan), 1.00, fac=1.10)
    an_lim = numpy.power(10, an_lim)
    dan_lim = numpy.power(10, dan_lim)

    gmr_lim = None if gmr is None else growth_lim(gmr, 0.95, fac=1.10)
    t_lim = growth_lim(numpy.ma.log10(t), 0.99, fac=1.10)
    dt_lim = growth_lim(numpy.ma.log10(dt), 0.99, fac=1.10)
    tn_lim = growth_lim(numpy.ma.log10(tn), 1.00, fac=1.10)
    dtn_lim = growth_lim(numpy.ma.log10(dtn), 1.00, fac=1.10)
    dtn_lim = numpy.power(10, dtn_lim)
    tn_lim = numpy.power(10, tn_lim)

    sv_lim = growth_lim(svel, 0.90, fac=1.2, midpoint=0)
    so_lim = numpy.power(10., growth_lim(numpy.ma.log10(ssigo), 1.00, fac=1.05))
    sc_lim = growth_lim(ssigcor, 0.90, fac=1.05)
    ss_lim = numpy.power(10., growth_lim(numpy.ma.log10(ssigc), 0.90, fac=1.05))

#    so_lim = [ min(so_lim[0],ss_lim[0]), max(so_lim[1],ss_lim[1]) ]
#    ss_lim = so_lim.copy()


    uniq, indx = numpy.unique(binid, return_index=True)
    if uniq[0] == -1:
        indx = indx[1:]

#    fig = pyplot.figure()
#    ax = fig.add_subplot(111, projection='3d')
#
#    ax.scatter(numpy.ma.log10(dtn[indx]), numpy.ma.log10(dan[indx]), numpy.ma.log10(ssigo[indx]),
#               c=snr[indx], cmap='viridis', norm=colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1]))
#
#    ax.set_xlim(numpy.log10(dtn_lim))
#    ax.set_ylim(numpy.log10(dan_lim))
#    ax.set_zlim(numpy.log10(so_lim))


#    sc = pyplot.scatter(dan[indx], ssigo[indx], marker='.', s=20, c=rchi2[indx],
#                        cmap='viridis', vmin=chi_lim[0], vmax=chi_lim[1])
#    sc = pyplot.scatter(snr[indx], ssigo[indx], marker='.', s=20, c=dan[indx],
#                        cmap='viridis', norm=colors.LogNorm(vmin=numpy.amin(dan[indx]), vmax=numpy.amax(dan[indx])))
#    sc = pyplot.scatter(signal[indx], dan[indx], marker='.', s=20, c=snr[indx],
#                        cmap='viridis', norm=colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1]))
    sc = pyplot.scatter(snr[indx], dan[indx], marker='.', s=20, c=ssigo[indx],
                        cmap='viridis', norm=colors.LogNorm(vmin=so_lim[0], vmax=so_lim[1]))
#    sc = pyplot.scatter(dan[indx], ssigo[indx], marker='.', s=20, c=snr[indx],
#                        cmap='viridis', norm=colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1]))
#    sc = pyplot.scatter(dtn[indx], dan[indx], marker='.', s=20, c=snr[indx],
#                        cmap='viridis', norm=colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1]))
    cb = pyplot.colorbar(sc)
##    pyplot.xlim(s_lim)
    pyplot.ylim(dan_lim)
    pyplot.xlim(snr_lim)
    pyplot.xscale('log')
    pyplot.yscale('log')

    pyplot.show()
    exit()
    

def masked_imshow(fig, ax, cax, data, extent=None, norm=None, vmin=None, vmax=None, cmap='viridis',
                  zorder=0, cbformat=None, subs=None):
    if numpy.sum(data.mask) != numpy.prod(data.shape):
        if norm is None:
            img = ax.imshow(data, origin='lower', interpolation='nearest', extent=extent,
                            vmin=vmin, vmax=vmax, cmap=cmap, zorder=zorder)
        else:
            img = ax.imshow(data, origin='lower', interpolation='nearest', extent=extent,
                            norm=norm, cmap=cmap, zorder=zorder)

        cb = fig.colorbar(img, cax=cax) if cbformat is None else \
                    fig.colorbar(img, cax=cax, format=cbformat)
        if subs is not None:
            cb.locator = ticker.LogLocator(base=10, subs=(1.,2.,4.,))
            cb.update_ticks()
    else:
        _norm = colors.Normalize(vmin=vmin, vmax=vmax) if norm is None else norm
        cb = colorbar.ColorbarBase(cax, cmap=cm.get_cmap(cmap), norm=_norm)

# TODO:
#   - Add a buffer keyword that increases the size of the image panels
#   - Use astropy image plotting (with wcs) tools instead?
def stellar_continuum_maps(name, snr, r68, r99, rchi2, signal, a, da, an, dan,
                           gmr, t, dt, tn, dtn, svel, ssigo, ssigcor, ssigc, extent=None,
                           fwhm=2.5, ofile=None):

    font = { 'size' : 6 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(2*w,2*h))

    x_order = 1 if extent[0] < extent[1] else -1
    Dx = x_order*(extent[1] - extent[0])
    y_order = 1 if extent[2] < extent[3] else -1
    Dy = y_order*(extent[3] - extent[2])
    # Force the image panels to be square; assumes extent has the same units in
    # both dimensions.
    D = max(Dx, Dy)

    x_lim = (extent[0]+extent[1])/2 + D * x_order * numpy.array([-1,1])/2
    y_lim = (extent[2]+extent[3])/2 + D * y_order * numpy.array([-1,1])/2

    snr_lim = numpy.power(10., growth_lim(numpy.ma.log10(snr), 0.90, fac=1.05))
    snr_lim[0] = max(0.1, snr_lim[0])
    r68_lim = numpy.power(10., growth_lim(numpy.ma.log10(r68), 0.90, fac=1.05))
    r99_lim = numpy.power(10., growth_lim(numpy.ma.log10(r99), 0.90, fac=1.05))
    chi_lim = growth_lim(rchi2, 0.90, fac=1.05)

    s_lim = numpy.power(10., growth_lim(numpy.ma.log10(signal), 0.99, fac=1.05))
    a_lim = growth_lim(numpy.ma.log10(a), 0.99, fac=1.05)
    da_lim = growth_lim(numpy.ma.log10(da), 0.99, fac=1.05)
    an_lim = growth_lim(numpy.ma.log10(an), 0.99, fac=1.05)
    dan_lim = growth_lim(numpy.ma.log10(dan), 0.99, fac=1.05)

    gmr_lim = None if gmr is None else growth_lim(gmr, 0.95, fac=1.10)
    t_lim = growth_lim(numpy.ma.log10(t), 0.99, fac=1.05)
    dt_lim = growth_lim(numpy.ma.log10(dt), 0.99, fac=1.05)
    tn_lim = growth_lim(numpy.ma.log10(tn), 0.99, fac=1.05)
    dtn_lim = growth_lim(numpy.ma.log10(dtn), 0.99, fac=1.05)

    sv_lim = growth_lim(svel, 0.90, fac=1.2, midpoint=0)
    so_lim = numpy.power(10., growth_lim(numpy.ma.log10(ssigo), 0.90, fac=1.10))
    sc_lim = growth_lim(ssigcor, 0.90, fac=1.10)
    ss_lim = numpy.power(10., growth_lim(numpy.ma.log10(ssigc), 0.90, fac=1.10))

    so_lim = [ min(so_lim[0],ss_lim[0]), max(so_lim[1],ss_lim[1]) ]
    ss_lim = so_lim.copy()

#    so_lim = [10,400]
#    ss_lim = [10,400]

    # ------------------------------------------------------------------
    # Layout
    left = 0.035
    bott = 0.08
    imwd = 0.18
    hbuf = 0.06
    cbuf = 0.005
    cbwd = 0.01
    vbuf = 0.03
    # ------------------------------------------------------------------

    # ------------------------------------------------------------------
    # S/N
    i,j=0,3
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
#    print(numpy.sum(snr.mask), numpy.prod(snr.shape), numpy.sum(snr.mask)/numpy.prod(snr.shape))
    masked_imshow(fig, ax, cax, snr, extent=extent,
                  norm=colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1]), cmap='viridis',
                  zorder=3, cbformat='%.1f', subs=(1.,2.,4.))
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'S/N', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

    # ------------------------------------------------------------------
    # 68th percentile of fractional residuals
    i,j=1,3
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
#    print(numpy.sum(r68.mask), numpy.prod(r68.shape), numpy.sum(r68.mask)/numpy.prod(r68.shape))
    masked_imshow(fig, ax, cax, r68, extent=extent,
                  norm=colors.LogNorm(vmin=r68_lim[0], vmax=r68_lim[1]), cmap='viridis_r',
                  zorder=3, cbformat='%.2f', subs=(1.,2.,4.))
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'68% Frac. Resid', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

#    ax.text(0.5, 1.2, name, horizontalalignment='center',
#            verticalalignment='center', transform=ax.transAxes, fontsize=12)
#    ax.text(0.5, 1.08, method, horizontalalignment='center',
#            verticalalignment='center', transform=ax.transAxes, fontsize=12)

    # ------------------------------------------------------------------
    # 99th percentile of fractional residuals
    i,j=2,3
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
#    print(numpy.sum(r99.mask), numpy.prod(r99.shape), numpy.sum(r99.mask)/numpy.prod(r99.shape))
    masked_imshow(fig, ax, cax, r99, extent=extent,
                  norm=colors.LogNorm(vmin=r99_lim[0], vmax=r99_lim[1]), cmap='viridis_r',
                  zorder=3, cbformat='%.2f', subs=(1.,2.,4.))
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'99% Frac. Resid', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Reduced chi-square
    i,j=3,3
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
#    print(numpy.sum(rchi2.mask), numpy.prod(rchi2.shape),
#          numpy.sum(rchi2.mask)/numpy.prod(rchi2.shape))
    masked_imshow(fig, ax, cax, rchi2, extent=extent, vmin=chi_lim[0], vmax=chi_lim[1],
                  cmap='viridis_r', zorder=3, cbformat='%.2f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\chi^2_\nu$', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)


    # ------------------------------------------------------------------
    # Signal
    i,j=0,2
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
#    print(numpy.sum(signal.mask), numpy.prod(signal.shape),
#          numpy.sum(signal.mask)/numpy.prod(signal.shape))
    masked_imshow(fig, ax, cax, signal, extent=extent,
                  norm=colors.LogNorm(vmin=s_lim[0], vmax=s_lim[1]), cmap='viridis',
                  zorder=3, cbformat='%.1f', subs=(1.,2.,4.))
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$I_g$', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

    # ------------------------------------------------------------------
    # A
    i,j=1,2
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    lga = numpy.ma.log10(a)
#    print(numpy.sum(lga.mask), numpy.prod(lga.shape), numpy.sum(lga.mask)/numpy.prod(lga.shape))
    masked_imshow(fig, ax, cax, lga, extent=extent, vmin=a_lim[0], vmax=a_lim[1], cmap='viridis',
                  zorder=3, cbformat='%.1f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\log A$', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Delta A
    i,j=2,2
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    lgan = numpy.ma.log10(an)
#    print(numpy.sum(lgan.mask), numpy.prod(lgan.shape), numpy.sum(lgan.mask)/numpy.prod(lgan.shape))
    masked_imshow(fig, ax, cax, lgan, extent=extent, vmin=an_lim[0], vmax=an_lim[1],
                  cmap='viridis_r', zorder=3, cbformat='%.1f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\log A_n$', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Delta A/Ig
    i,j=3,2
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    lgdan = numpy.ma.log10(dan)
#    print(numpy.sum(lgdan.mask), numpy.prod(lgdan.shape),
#          numpy.sum(lgdan.mask)/numpy.prod(lgdan.shape))
    masked_imshow(fig, ax, cax, lgdan, extent=extent, vmin=dan_lim[0], vmax=dan_lim[1],
                  cmap='viridis_r', zorder=3, cbformat='%.1f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\log(\delta A_n)$', horizontalalignment='left',
            verticalalignment='center', transform=ax.transAxes)

    # ------------------------------------------------------------------
    # g-r Color
    i,j=0,1
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
#    print(numpy.sum(gmr.mask), numpy.prod(gmr.shape), numpy.sum(gmr.mask)/numpy.prod(gmr.shape))
    if gmr is not None:
        cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
        masked_imshow(fig, ax, cax, gmr, extent=extent, vmin=gmr_lim[0], vmax=gmr_lim[1],
                      cmap='RdBu_r', zorder=3, cbformat='%.1f')
    else:
        ax.text(0.5, 0.5, 'No g-r data', ha='center', va='center', transform=ax.transAxes)
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$g-r$', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

    # ------------------------------------------------------------------
    # T
    i,j=1,1
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    lgt = numpy.ma.log10(t)
#    print(numpy.sum(lgt.mask), numpy.prod(lgt.shape), numpy.sum(lgt.mask)/numpy.prod(lgt.shape))
    masked_imshow(fig, ax, cax, lgt, extent=extent, vmin=t_lim[0], vmax=t_lim[1],
                  cmap='viridis', zorder=3, cbformat='%.1f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\log T$', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Delta T
    i,j=2,1
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    lgtn = numpy.ma.log10(tn)
#    print(numpy.sum(lgtn.mask), numpy.prod(lgtn.shape), numpy.sum(lgtn.mask)/numpy.prod(lgtn.shape))
    masked_imshow(fig, ax, cax, lgtn, extent=extent, vmin=tn_lim[0], vmax=tn_lim[1],
                  cmap='viridis_r', zorder=3, cbformat='%.1f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\log T_n$', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Delta T/Ig
    i,j=3,1
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    lgdtn = numpy.ma.log10(dtn)
#    print(numpy.sum(lgdtn.mask), numpy.prod(lgdtn.shape),
#            numpy.sum(lgdtn.mask)/numpy.prod(lgdtn.shape))
    masked_imshow(fig, ax, cax, lgdtn, extent=extent, vmin=dtn_lim[0], vmax=dtn_lim[1],
                  cmap='viridis_r', zorder=3, cbformat='%.1f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\log \delta T_n$', horizontalalignment='left',
            verticalalignment='center', transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Stellar velocity
    i,j=0,0
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
#    print(numpy.sum(svel.mask), numpy.prod(svel.shape), numpy.sum(svel.mask)/numpy.prod(svel.shape))
    masked_imshow(fig, ax, cax, svel, extent=extent, vmin=sv_lim[0], vmax=sv_lim[1],
                  cmap='RdBu_r', zorder=3, cbformat='%.0f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$V_\ast$ (km/s)', horizontalalignment='left',
            verticalalignment='center', transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Stellar sigma
    i,j=1,0
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
#    print(numpy.sum(ssigo.mask), numpy.prod(ssigo.shape),
#            numpy.sum(ssigo.mask)/numpy.prod(ssigo.shape))
    masked_imshow(fig, ax, cax, ssigo, extent=extent,
                  norm=colors.LogNorm(vmin=so_lim[0], vmax=so_lim[1]),
                  cmap='viridis', zorder=3, cbformat='%.0f', subs=(1.,2.,4.))
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\sigma_{\rm obs}$ (km/s)', horizontalalignment='left',
            verticalalignment='center', transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Sigma correction
    i,j=2,0
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
#    print(numpy.sum(ssigcor.mask), numpy.prod(ssigcor.shape),
#            numpy.sum(ssigcor.mask)/numpy.prod(ssigcor.shape))
    masked_imshow(fig, ax, cax, ssigcor, extent=extent, vmin=sc_lim[0], vmax=sc_lim[1],
                  cmap='viridis', zorder=3, cbformat='%.0f')
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\sigma_{\rm corr}$ (km/s)', horizontalalignment='left',
            verticalalignment='center', transform=ax.transAxes)

    # ------------------------------------------------------------------
    # Corrected stellar sigma 
    i,j=3,0
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
#    print(numpy.sum(ssigc.mask), numpy.prod(ssigc.shape),
#            numpy.sum(ssigc.mask)/numpy.prod(ssigc.shape))
    masked_imshow(fig, ax, cax, ssigo, extent=extent,
                  norm=colors.LogNorm(vmin=ss_lim[0], vmax=ss_lim[1]),
                  cmap='viridis', zorder=3, cbformat='%.0f', subs=(1.,2.,4.))
    ax.set_xlim(x_lim) #extent[:2])
    ax.set_ylim(y_lim) #extent[2:])
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.1, 0.95, r'$\sigma_\ast$ (km/s)', horizontalalignment='left',
            verticalalignment='center', transform=ax.transAxes)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)


# TODO: Should make this a datacube member function
def gmr_data(cube, min_frac=0.8):
    """
    Read or construct a g-r for the provided datacube.
    """
    # First see if there are GIMG and RIMG extensions in the input datacube
    # (requires a MaNGA DRP-like datamodel)
    with fits.open(str(cube.file_path)) as hdu:
        ext_list = [h.name for h in hdu]
        if 'GIMG' in ext_list and 'RIMG' in ext_list:
            return -2.5*numpy.ma.log10(numpy.ma.MaskedArray(hdu['GIMG'].data,
                                                            mask=numpy.invert(hdu['GIMG'].data>0))
                                    / numpy.ma.MaskedArray(hdu['RIMG'].data,
                                                           mask=numpy.invert(hdu['RIMG'].data>0)))

    # Next, try to create it from scratch
    g_file = defaults.dap_data_root() / 'filter_response' / 'gunn_2001_g_response.db'
    if not g_file.exists():
        warnings.warn(f'Could not find file {g_file}; cannot produce rough g-r image.')
        return None

    r_file = defaults.dap_data_root() / 'filter_response' / 'gunn_2001_r_response.db'
    if not r_file.exists():
        warnings.warn(f'Could not find file {r_file}; cannot produce rough g-r image.')
        return None

    flux = cube.copy_to_masked_array()
    # Fix the ordering
    flux = flux.reshape(cube.flux.shape).T.reshape(npix,-1)
    gpm = numpy.logical_not(numpy.ma.getmaskarray(flux)).astype(float)

    # Wavelengths in these files are in angstroms
    g_wave, g_eff = numpy.genfromtxt(str(g_file)).T
    indx = g_eff > 0.
    g_lo = numpy.amin(g_wave[indx])
    g_hi = numpy.amax(g_wave[indx])
    g_wgt = interpolate.interp1d(g_wave[indx], g_eff[indx], bounds_error=False,
                                 fill_value=0.)(cube.wave)[None,:]*gpm
    g_mask = numpy.sum((g_wgt[:, 1:]>0).astype(float) * numpy.diff(cube.wave)[None, :],
                       axis=1).reshape(cube.spatial_shape[::-1]) < min_frac*(g_hi-g_lo)
    if numpy.all(g_mask):
        warnings.warn('Datacube has insufficient coverage of the g-band to construct a g-r image.')
        return None

    r_wave, r_eff = numpy.genfromtxt(str(r_file)).T
    indx = r_eff > 0.
    r_lo = numpy.amin(r_wave[indx])
    r_hi = numpy.amax(r_wave[indx])
    r_wgt = interpolate.interp1d(r_wave[indx], r_eff[indx], bounds_error=False,
                                 fill_value=0.)(cube.wave)[None,:]*gpm

    r_mask = numpy.sum((r_wgt[:, 1:]>0).astype(float) * numpy.diff(cube.wave)[None, :],
                       axis=1).reshape(cube.spatial_shape) < min_frac*(r_hi-r_lo)
    if numpy.all(r_mask):
        warnings.warn('Datacube has insufficient coverage of the r-band to construct a g-r image.')
        return None

    g = numpy.sum(g_wgt[:, 1:] * flux[:, 1:] * numpy.diff(cube.wave)[None,:],
                       axis=1).reshape(cube.spatial_shape[::-1])
    r = numpy.sum(r_wgt[:, 1:] * flux[:, 1:] * numpy.diff(cube.wave)[None,:],
                       axis=1).reshape(cube.spatial_shape[::-1])
    return -2.5*numpy.ma.log10(numpy.ma.divide(numpy.ma.MaskedArray(g, mask=g_mask),
                                               numpy.ma.MaskedArray(r, mask=r_mask)))


def rdxqa_data(cube, rdxqa_method, output_path):
    # Get the surface brightness and S/N maps from the
    # ReductionAssessments object

    directory, file = ReductionAssessment.default_paths(cube, rdxqa_method['key'],
                                                        output_path=output_path)
    rdxqa_file = directory / file
    if not rdxqa_file.exists():
        raise FileNotFoundError(f'{rdxqa_file} does not exist!')

    with fits.open(str(rdxqa_file)) as hdu:
        spatial_shape = (int(numpy.sqrt(hdu['SPECTRUM'].data['SNR'].size)),)*2
        fgood_map = hdu['SPECTRUM'].data['FGOODPIX'].reshape(spatial_shape).T
        signal_map = numpy.ma.MaskedArray(hdu['SPECTRUM'].data['SIGNAL'].reshape(spatial_shape).T,
                                          mask=numpy.invert(fgood_map > 0))
        snr_map = numpy.ma.MaskedArray(hdu['SPECTRUM'].data['SNR'].reshape(spatial_shape).T,
                                       mask=numpy.invert(fgood_map > 0))
    return signal_map, snr_map

    
def continuum_component_data(cube, rdxqa_method, binning_method, continuum_method, output_path,
                             signal_map=None, tpl_flux_renorm=None):

    # Get the coefficient data from the StellarContinuumModel object
    directory, file = StellarContinuumModel.default_paths(cube, continuum_method['key'],
                                                          rdxqa_method['key'],
                                                          binning_method['key'],
                                                          output_path=output_path)
    sc_file = directory / file
    if not sc_file.exists():
        raise FileNotFoundError(f'{sc_file} does not exist!')

    hdu = fits.open(str(sc_file))
    # This only has the bins with stellar continuum measurements
    binid_map = hdu['BINID'].data.copy()

    if signal_map is None:
        signal = numpy.ones(hdu['PAR'].data['BINID_INDEX'].size, dtype=float)
    else:
        # Find the unique indices in the map with valid bins
        unique, unique_indx = numpy.unique(binid_map.ravel(), return_index=True)
        if unique[0] == -1:
            unique_indx = unique_indx[1:]
        # .. and pull these out of the signal map
        signal = signal_map.ravel()[unique_indx]

#    print(hdu['PAR'].data['TPLWGT'].shape)
    w = hdu['PAR'].data['TPLWGT'].copy()
    if tpl_flux_renorm is not None:
        w /= tpl_flux_renorm[None,:]
    wn = w/signal[:,None]
#    t = numpy.sqrt(numpy.sum(numpy.square(w), axis=1))
    t = numpy.sum(numpy.absolute(w), axis=1)
    tn = numpy.sum(numpy.absolute(wn), axis=1)
#    dtw = numpy.median(w / numpy.sum(w, axis=1)[:,None], axis=0)
    dtw = numpy.ones(w.shape[1], dtype=float)
#    dtnw = numpy.median(wn / numpy.sum(wn, axis=1)[:,None], axis=0)
    dtnw = numpy.ones(w.shape[1], dtype=float)
    dt = numpy.ma.sqrt(numpy.sum(dtw[None,:]*numpy.square(w - numpy.ma.median(w, axis=0)[None,:]),
                            axis=1) / numpy.sum(dtw)).filled(0.0)
    dtn = numpy.ma.sqrt(numpy.sum(dtnw[None,:]*numpy.square(wn - numpy.ma.median(wn, axis=0)[None,:]),   
                            axis=1) / numpy.sum(dtnw)).filled(0.0)

    w = hdu['PAR'].data['ADDCOEF'].copy()
    wn = w/signal[:,None]
#    a = numpy.sqrt(numpy.sum(numpy.square(w), axis=1))
    a = numpy.sum(numpy.absolute(w), axis=1)
    an = numpy.sum(numpy.absolute(wn), axis=1)
#    daw = numpy.median(numpy.absolute(w)
#                        / numpy.sum(numpy.absolute(w), axis=1)[:,None], axis=0)
    daw = numpy.ones(w.shape[1], dtype=float)
#    danw = numpy.median(numpy.absolute(wn)
#                        / numpy.sum(numpy.absolute(wn), axis=1)[:,None], axis=0)
    danw = numpy.ones(w.shape[1], dtype=float)
    da = numpy.ma.sqrt(numpy.sum(daw[None,:]*numpy.square(w - numpy.ma.median(w, axis=0)[None,:]),
                            axis=1) / numpy.sum(daw)).filled(0.0)
    dan = numpy.ma.sqrt(numpy.sum(danw[None,:]*numpy.square(wn - numpy.ma.median(wn, axis=0)[None,:]),
                            axis=1) / numpy.sum(danw)).filled(0.0)

    hdu.close()
    del hdu

    return binid_map, t, dt, tn, dtn, a, da, an, dan

    
def maps_data(maps_file):
    if not maps_file.exists():
        raise FileNotFoundError(f'{maps_file} does not exist!')

    with fits.open(str(maps_file)) as hdu:

        mask = hdu['BINID'].data[1,:,:] < 0
        r68_map = numpy.ma.MaskedArray(hdu['STELLAR_FOM'].data[3,:,:], mask=mask)
        r99_map = numpy.ma.MaskedArray(hdu['STELLAR_FOM'].data[4,:,:], mask=mask)
        rchi2_map = numpy.ma.MaskedArray(hdu['STELLAR_FOM'].data[2,:,:], mask=mask)
        svel_map = numpy.ma.MaskedArray(hdu['STELLAR_VEL'].data[:,:], mask=mask)
        ssigo_map = numpy.ma.MaskedArray(hdu['STELLAR_SIGMA'].data[:,:], mask=mask)
        ssigcor_map = numpy.ma.MaskedArray(hdu['STELLAR_SIGMACORR'].data[1,:,:], mask=mask)
        ssigc_map = numpy.ma.sqrt( numpy.square(ssigo_map) - numpy.square(ssigcor_map) )

        extent = map_extent(hdu, 'SPX_MFLUX')

    return r68_map, r99_map, rchi2_map, svel_map, ssigo_map, ssigcor_map, ssigc_map, extent


def ppxffit_qa_plot(cube, plan, method_dir, ref_dir, qa_dir, fwhm=None, tpl_flux_renorm=None):
    """
    Construct a QA plot for the PPXFFit results.

    Args:
        cube (:class:`~mangadap.datacube.datacube.DataCube`):
            Analyzed data cube
        plan (:obj:`dict`, :class:`~mangadap.par.parset.ParSet`):
            Object with analysis plan keywords.
        method_dir (`Path`_):
            Top-level "method" directory with all the DAP files.
        ref_dir (`Path`_):
            Directory with all the DAP reference files.
        qa_dir (`Path`_):
            Directory for the QA figures
        tpl_flux_renorm (`numpy.ndarray`_, optional):
            Renormalization coefficients for all the templates.  Shape must
            match the number of templates for the stellar conntinuum fit.
    """
    # Check that the paths exists
    if not method_dir.exists():
        raise NotADirectoryError(f'{method_dir} does not exist; run the DAP first!')
    if not ref_dir.exists():
        raise NotADirectoryError(f'{ref_dir} does not exist; run the DAP first!')
    if not qa_dir.exists():
        qa_dir.mkdir(parents=True)

    # Get the name of the output file
    ofile = qa_dir / f'{cube.output_root}-MAPS-{plan["key"]}-ppxffit.png'

    # Get the g-r map
    gmr_map = gmr_data(cube)

    # Get the reduction assessment data
    rdxqa_method = ReductionAssessmentDef.from_dict(plan['rdxqa'])
    signal_map, snr_map = rdxqa_data(cube, rdxqa_method, ref_dir)

    # Get the template weights and additive-polynomial coefficients
    binning_method = SpatiallyBinnedSpectraDef.from_dict(plan['binning'])
    continuum_method = StellarContinuumModelDef.from_dict(plan['continuum'])
    binid_map, t, dt, tn, dtn, a, da, an, dan \
            = continuum_component_data(cube, rdxqa_method, binning_method, continuum_method,
                                       ref_dir, signal_map=signal_map,
                                       tpl_flux_renorm=tpl_flux_renorm)

#    print('t:', numpy.sum(t > 0), numpy.prod(t.shape))
#    print('dt:', numpy.sum(dt > 0), numpy.prod(dt.shape))
#    print('tn:', numpy.sum(tn > 0), numpy.prod(tn.shape))
#    print('dtn:', numpy.sum(dtn > 0), numpy.prod(dtn.shape))

#    print('a:', numpy.sum(a > 0), numpy.prod(a.shape))
#    print('da:', numpy.sum(da > 0), numpy.prod(da.shape))
#    print('an:', numpy.sum(an > 0), numpy.prod(an.shape))
#    print('dan:', numpy.sum(dan > 0), numpy.prod(dan.shape))

    # Map the continuum component data
    t_map = DAPFitsUtil.reconstruct_map(binid_map.shape, binid_map.ravel(), t, quiet=True)
    t_map = numpy.ma.MaskedArray(t_map, mask=numpy.invert(t_map > 0))

    dt_map = DAPFitsUtil.reconstruct_map(binid_map.shape, binid_map.ravel(), dt, quiet=True)
    dt_map = numpy.ma.MaskedArray(dt_map, mask=numpy.invert(dt_map > 0)) #/t_map

    tn_map = DAPFitsUtil.reconstruct_map(binid_map.shape, binid_map.ravel(), tn, quiet=True)
    tn_map = numpy.ma.MaskedArray(tn_map, mask=numpy.invert(tn_map > 0))

    dtn_map = DAPFitsUtil.reconstruct_map(binid_map.shape, binid_map.ravel(), dtn, quiet=True)
    dtn_map = numpy.ma.MaskedArray(dtn_map, mask=numpy.invert(dtn_map > 0)) #/tn_map

    a_map = DAPFitsUtil.reconstruct_map(binid_map.shape, binid_map.ravel(), a, quiet=True)
    a_map = numpy.ma.MaskedArray(a_map, mask=numpy.invert(a_map > 0))

    da_map = DAPFitsUtil.reconstruct_map(binid_map.shape, binid_map.ravel(), da, quiet=True)
    da_map = numpy.ma.MaskedArray(da_map, mask=numpy.invert(da_map > 0)) #/a_map

    an_map = DAPFitsUtil.reconstruct_map(binid_map.shape, binid_map.ravel(), an, quiet=True)
    an_map = numpy.ma.MaskedArray(an_map, mask=numpy.invert(an_map > 0))

    dan_map = DAPFitsUtil.reconstruct_map(binid_map.shape, binid_map.ravel(), dan, quiet=True)
    dan_map = numpy.ma.MaskedArray(dan_map, mask=numpy.invert(dan_map > 0)) #/an_map

    _, directory, file = dapfits.default_paths(cube, method=plan['key'], output_path=method_dir)
    maps_file = directory / file
    
    # Get the remaining output maps from the MAPS file
    r68_map, r99_map, rchi2_map, svel_map, ssigo_map, ssigcor_map, ssigc_map, extent \
            = maps_data(maps_file)

    stellar_continuum_maps(cube.name, snr_map, r68_map, r99_map, rchi2_map, signal_map,
                           a_map, da_map, an_map, dan_map, gmr_map, t_map, dt_map, tn_map, dtn_map,
                           svel_map, ssigo_map, ssigcor_map, ssigc_map, fwhm=fwhm, extent=extent,
                           ofile=ofile)


class PpxfFitQA(scriptbase.ScriptBase):

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Construct QA plots for pPXF fit.', width=width)

        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('-c', '--config', type=str, default=None,
                           help='Configuration file used to instantiate the relevant DataCube '
                                'derived class.')
        group.add_argument('-f', '--cubefile', type=str, default=None,
                           help='Name of the file with the datacube data.  Must be possible to '
                                'instantiate the relevant DataCube derived class directly from '
                                'the file only.')
        parser.add_argument('--cube_module', nargs='*',
                            default='mangadap.datacube.MaNGADataCube',
                            help='The name of the module that contains the DataCube derived '
                                 'class used to read the data.')

        parser.add_argument('-p', '--plan', type=str,
                            help='SDSS parameter file with analysis plan.  If not provided, a '
                                 'default plan is used.')
        parser.add_argument('--plan_module', nargs='*',
                            default='mangadap.config.manga.MaNGAAnalysisPlan',
                            help='The name of the module used to define the analysis plan and '
                                 'the output paths.')

        parser.add_argument('-o', '--output_path', type=str, default=None,
                            help='Top-level directory for the DAP output files; default path is '
                                 'set by the provided analysis plan object (see plan_module).')

        # TODO: For MaNGA, pull fwhm from the header.  Put FWHM in cube.meta?
        parser.add_argument('-b', '--beam', type=float, default=None, help='Beam FWHM for plot.')

        parser.add_argument('--normal_backend', dest='bgagg', action='store_false', default=True)

        parser.add_argument('--template_flux_file', type=str, default=None,
                            help='template renormalization flux file.  Will attempt to read '
                                 'default if not provided.  If no file is provided and the '
                                 'default file does not exist, no renormalization of the '
                                 'templates is performed.')
        return parser

    @staticmethod
    def main(args):
        t = time.perf_counter()

        if args.bgagg:
            pyplot.switch_backend('agg')

        # Instantiate the DataCube
        #   - Import the module used to read the datacube
        if isinstance(args.cube_module, list) and len(args.cube_module) > 2:
            raise ValueError('Provided cube module must be one or two strings.')
        if isinstance(args.cube_module, str) or len(args.cube_module) == 1:
            UserDataCube = load_object(args.cube_module if isinstance(args.cube_module, str) 
                                       else args.cube_module[0])
        else:
            UserDataCube = load_object(args.cube_module[0], obj=args.cube_module[1])
        #   - Check that the class is derived from DataCube
        if not issubclass(UserDataCube, DataCube):
            raise TypeError('Defined cube object must subclass from mangadap.datacube.DataCube.')

        #   - Import the module used to set the analysis plan
        if isinstance(args.plan_module, list) and len(args.plan_module) > 2:
            raise ValueError('Provided plan module must be one or two strings.')
        if isinstance(args.plan_module, str) or len(args.plan_module) == 1:
            UserPlan = load_object(args.plan_module if isinstance(args.plan_module, str) 
                                       else args.plan_module[0])
        else:
            UserPlan = load_object(args.plan_module[0], obj=args.plan_module[1])
        #   - Check that the class is derived from AnalysisPlan
        if not issubclass(UserPlan, AnalysisPlan):
            raise TypeError('Defined plan object must subclass from '
                            'mangadap.config.analysisplan.AnalysisPlan')

        #   - Instantiate using either the datacube file directly or a
        #     configuration file
        cube = UserDataCube(args.cubefile) if args.config is None \
                    else UserDataCube.from_config(args.config)

        # Read the analysis plan
        plan = UserPlan.default(cube=cube, analysis_path=args.output_path) if args.plan is None \
                    else UserPlan.from_toml(args.plan, cube=cube, analysis_path=args.output_path)

        # Construct the plot for each analysis plan
        for i, key in enumerate(plan.keys()):

            # Get the template library keyword
            if args.template_flux_file is None:
                sc_method = StellarContinuumModelDef.from_dict(plan[key]['continuum'])
                tpl_key = sc_method['fitpar']['template_library']['key']
                library_root = sc_method["fitpar"]["template_library"].get_file_list()[0].parent
                tpl_renorm_file = library_root / f'{tpl_key.lower()}_fluxes.db'
                #sc_method = StellarContinuumModel.define_method(plan[i]['continuum_key'])
                #tpl_key = sc_method['fitpar']['template_library_key']
                #library = TemplateLibrary.define_library(tpl_key)
                #library_root = library.get_file_list()[0].parent
                #tpl_renorm_file = library_root / f'{tpl_key.lower()}_fluxes.db'
            else:
                tpl_renorm_file = Path(args.template_flux_file).resolve()

            if not tpl_renorm_file.exists():
                warnings.warn('Could not find file with template renormalizations: '
                                f'{tpl_renorm_file}')
                tpl_renorm_file = None

            tpl_flux_renorm = None if tpl_renorm_file is None \
                                else numpy.genfromtxt(tpl_renorm_file, dtype=float)[:,2]

            # Construct the plot
            ppxffit_qa_plot(cube, plan[key], plan.method_path(plan_index=i),
                            plan.method_path(plan_index=i, ref=True),
                            plan.method_path(plan_index=i, qa=True), fwhm=args.beam,
                            tpl_flux_renorm=tpl_flux_renorm)

        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))






import os
import time
import argparse

import numpy

from astropy.io import fits
import astropy.constants

from mangadap.datacube import MaNGADataCube
from mangadap.survey.drpcomplete import DRPComplete

from mangadap.scripts import scriptbase


class WriteDapConfig(scriptbase.ScriptBase):

    @classmethod
    def name(cls):
        """
        Return the name of the executable.
        """
        return 'write_dap_config'

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Generate a DAP input configuration file',
                                    width=width)

        parser.add_argument('plate', type=int, help='Plate number')
        parser.add_argument('ifudesign', type=int, help='IFU design number')
        parser.add_argument('ofile', type=str, help='Output file name')
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('-c', '--drpcomplete', type=str, help='DRP complete fits file',
                            default=None)
        group.add_argument('-a', '--drpall', type=str, help='DRPall fits file', default=None)

        parser.add_argument('--sres_ext', type=str, default=None,
                            help='Spectral resolution extension to use.  Default set by '
                                 'MaNGADataCube class.')
        parser.add_argument('--sres_fill', type=str, default=None,
                            help='If present, use interpolation to fill any masked pixels in the '
                                 'spectral resolution vectors. Default set by MaNGADataCube class.')
        parser.add_argument('--covar_ext', type=str, default=None,
                            help='Use this extension to define the spatial correlation matrix.  '
                                 'Default set by MaNGADataCube class.')
        parser.add_argument('--drpver', type=str, default=None,
                            help='DRP version.  Default set by MaNGADataCube class.')
        parser.add_argument('--redux_path', type=str, default=None,
                            help='Path to the top-level DRP reduction directory.  Default set by '
                                'MaNGADataCube class.')
        parser.add_argument('--directory_path', type=str, default=None,
                            help='Exact path to the directory with the MaNGA DRP datacube.  The '
                                 'name of the file itself must match the nominal MaNGA DRP naming '
                                 'convention.  Default set by MaNGADataCube class.')
        parser.add_argument('-o', '--overwrite', default=False, action='store_true',
                            help='Overwrite any existing files.')
        return parser

    @staticmethod
    def main(args):
        t = time.perf_counter()

        if args.drpcomplete is not None:
            # Use the DRPcomplete file
            root_dir = os.path.dirname(args.drpcomplete)
            if len(root_dir) == 0:
                root_dir = '.'
            drpver = args.drpcomplete[args.drpcomplete.find('_v')+1:args.drpcomplete.find('.fits')]
            drpc = DRPComplete(drpver=drpver, directory_path=root_dir, readonly=True)
            index = drpc.entry_index(args.plate, args.ifudesign)
            MaNGADataCube.write_config(args.ofile, drpc['PLATE'][index], drpc['IFUDESIGN'][index],
                                       log=True,
                                       z=drpc['VEL'][index]/astropy.constants.c.to('km/s').value,
                                       vdisp=drpc['VDISP'][index], ell=drpc['ELL'][index],
                                       pa=drpc['PA'][index], reff=drpc['REFF'][index],
                                       sres_ext=args.sres_ext, sres_fill=args.sres_fill,
                                       covar_ext=args.covar_ext, drpver=args.drpver,
                                       redux_path=args.redux_path, overwrite=args.overwrite)
            return

        # Use the DRPall file
        with fits.open(args.drpall) as hdu:
            indx = numpy.where(hdu['MANGA'].data['PLATEIFU'] == '{0}-{1}'.format(args.plate,
                                                                                args.ifudesign))[0]
            if len(indx) != 1:
                raise ValueError('{0}-{1} either does not exist or has more than one match!'.format(
                                    args.plate, args.ifudesign))

            MaNGADataCube.write_config(args.ofile, args.plate, args.ifudesign,
                                       z=hdu[1].data['z'][indx[0]],
                                       ell=1-hdu[1].data['nsa_elpetro_ba'][indx[0]],
                                       pa=hdu[1].data['nsa_elpetro_phi'][indx[0]],
                                       reff=hdu[1].data['nsa_elpetro_th50_r'][indx[0]],
                                       sres_ext=args.sres_ext, sres_fill=args.sres_fill,
                                       covar_ext=args.covar_ext, drpver=args.drpver,
                                       redux_path=args.redux_path,
                                       directory_path=args.directory_path,
                                       overwrite=args.overwrite)

        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))




#!/usr/bin/env python3

import os
import time
import numpy

from matplotlib import pyplot, ticker, rc, colors

from astropy.io import fits
import astropy.constants

from mangadap.proc.util import growth_lim
from mangadap.util.bitmask import BitMask

#-----------------------------------------------------------------------------
def init_ax(fig, pos):
    ax = fig.add_axes(pos, facecolor='0.95')
    ax.grid(True, which='major', color='0.9', zorder=0, linestyle='-')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    return ax


def ew_d4000(drpall, dapall):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == 'HYB10-GAU-MILESHC')
    indx &= (drpall['nsa_sersic_mass'][dapall['DRPALLINDX']] > 0)

#    ha_indx = indx & (dapall['EMLINE_GSB_1RE'][:,18] > -999) \
#                    & (dapall['SPECINDEX_1RE'][:,44] > -999)
    ha_indx = indx & (dapall['EMLINE_GEW_1RE'][:,18] > -999) \
                    & (dapall['SPECINDEX_1RE'][:,44] > -999)
    hd_indx = indx & (dapall['SPECINDEX_1RE'][:,21] > -999) \
                    & (dapall['SPECINDEX_1RE'][:,44] > -999)

    mass_lim = [1e9, 7e11]
    d4_lim = [0.9, 2.3]
#    ha_lim = [0.005, 800]
    ha_lim = [2e-2, 400]
    hd_lim = [-4, 9]

    font = { 'size' : 12 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.3, 0.56, 0.4, 0.4])
    ax.set_xlim(d4_lim)
    ax.set_ylim(ha_lim)
    ax.tick_params(axis='x', which='both', direction='in')
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())

    sc = ax.scatter(dapall['SPECINDEX_1RE'][ha_indx,44], dapall['EMLINE_GEW_1RE'][ha_indx,18],
                    lw=0, marker='.', s=10, zorder=2,
                    c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][ha_indx]],
                    norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
                    cmap='viridis')

    ax.text(-0.25, 0.5, r'$\langle {\rm EW}_{{\rm H}\alpha}\rangle_e$ [$\AA$, in emission]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')

    ax = init_ax(fig, [0.3, 0.15, 0.4, 0.4])
    ax.set_xlim(d4_lim)
    ax.set_ylim(hd_lim)

    sc = ax.scatter(dapall['SPECINDEX_1RE'][hd_indx,44], dapall['SPECINDEX_1RE'][hd_indx,21],
                    lw=0, marker='.', s=10, zorder=2,
                    c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][hd_indx]],
                    norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
                    cmap='viridis')

    cax = fig.add_axes([0.51, 0.48, 0.15, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[1e9,1e10,1e11])
                         #, ticks=[0.1,1], format=ticker.ScalarFormatter())
#    pyplot.setp(cb.ax.xaxis.get_ticklabels(), fontsize=10)
    cax.text(0.5, 2.8, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
             ha='center', va='center', transform=cax.transAxes) #, fontsize=10)

    ax.text(-0.25, 0.5, r'$\langle {\rm H}\delta_{\rm A}\rangle_e$ [$\AA$, in absorption]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')

    ax.text(0.5, -0.18, r'$\langle {\rm Dn4000}\rangle_e$',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

#    ax.text(-0.22, 0.5, r'$\Delta V_\ast$ [km/s]',
#            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
#            rotation='vertical')
#    ax.text(0.55, 0.89, r'$\mathcal{M}_\ast\propto\ \Delta V^4$',
#            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
#            rotation=20, zorder=4)

    ofile = '../ms/figs/ew_d4000.pdf'
    fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)


#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.clock()

    dapall_file = '/Users/westfall/Work/MaNGA/testing/dr15/dapall-v2_4_3-2.2.1.fits'
    dapall = fits.open(dapall_file)['DAPALL'].data

    drpall_file = '/Users/westfall/Work/MaNGA/testing/dr15/drpall-v2_4_3.fits'
    drpall = fits.open(drpall_file)[1].data

    ew_d4000(drpall, dapall)

    print('Elapsed time: {0} seconds'.format(time.clock() - t))

